from __future__ import print_function
from collections import namedtuple, defaultdict
from argtools import command, argument
from ..utils import Counter
import re
import os
import io
import logging


class Contig(object):
    __slots__ = ('name_line', 'name', '_seqs')

    def __init__(self, name_line=''):
        assert not name_line.startswith('>'), 'please remove charcter ">"'
        self.name_line = name_line
        self.name = name_line.split(' ', 2)[0]
        self._seqs = []

    def append(self, seq):
        self._seqs.append(seq)

    def get_seq(self):
        return ''.join(self._seqs)

def iter_contigs(fp):
    contig = None
    try:
        while fp:
            line = next(fp).rstrip()
            if line.startswith('>'):
                if contig is not None:
                    yield contig
                contig = Contig(line[1:])
            else:
                contig.append(line)
    except StopIteration:
        if contig is not None:
            yield contig


@command.add_sub
@argument('fasta', nargs='?', default='/dev/stdin')
@argument('-w', '--width', type=int, default=50)
@argument('-u', '--uppercase', action='store_true')
def fa_fold(args):
    """ Set line breaks with fixed width
    """
    width = args.width
    with open(args.fasta) as fp:
        for contig in iter_contigs(fp):
            print ('>', contig.name_line, sep='')
            seq = contig.get_seq()
            if args.uppercase:
                seq = seq.upper()
            if not width:
                print (seq)
            else:
                reg = len(seq) // width
                last = len(seq) % width
                if reg:
                    for s in xrange(0, reg*width, width):
                        print (seq[s:s + width])
                if last:
                    print (seq[-last:])

@command.add_sub
@argument('fasta', nargs='?', default='/dev/stdin')
def fa_profile(args):
    """ Analyse contigs
    """
    print ('name', 'length', 'gc_ratio', 'at_ratio', 'count_A', 'count_C', 'count_G', 'count_T', 'count_N', sep='\t')
    with open(args.fasta) as fp:
        for contig in iter_contigs(fp):
            seq = contig.get_seq()
            length = len(seq)
            count = Counter(seq.upper())
            cA = count['A']
            cC = count['C']
            cG = count['G']
            cT = count['T']
            cN = count['N']
            gc_ratio = 1. * (cC + cG) / length
            at_ratio = 1. * (cA + cT) / length
            print (contig.name, length,
                   '{0:.3f}'.format(gc_ratio),
                   '{0:.3f}'.format(at_ratio),
                   cA, cC, cG, cT, cN, sep='\t')


@command.add_sub
@argument('fasta', nargs='?', default='/dev/stdin')
@argument('-m', '--min-run', type=int, default=4)
def fa_stats_hompol(args):
    """ Analyse contigs
    """
    print ('name', 'length', 'min_run',
           'count', 'max_len', 'min_len',
           'len_tot', 'len_AT', 'len_CG', 'len_A', 'len_T', 'len_C', 'len_G',
           'ratio', 'ratio_AT', 'ratio_CG', 'ratio_A', 'ratio_T', 'ratio_C', 'ratio_G',
           sep='\t')

    patterns = dict((base, '%s{%s,}' % (base, args.min_run)) for base in 'ACGT')
    pattern_res = dict((base, re.compile(patterns[base])) for base in 'ACGT')

    def get_ratio(a, b):
        return round(1. * a / b, 3)

    for contig in iter_contigs(open(args.fasta)):
        seq = contig.get_seq().upper()
        length = len(seq)
        base_lens = {}
        for base, reg in pattern_res.items():
            base_lens[base] = [m.end() - m.start() for m in reg.finditer(seq)]

        tot_len = sum(sum(base_lens[base]) for base in 'ACGT')
        count = sum(len(base_lens[base]) for base in 'ACGT')
        max_len = max(max(base_lens[base] or [0]) for base in 'ACGT')
        min_len = min(min(base_lens[base] or [float('inf')]) for base in 'ACGT')
        AT_len = sum(base_lens['A']) + sum(base_lens['T'])
        CG_len = sum(base_lens['C']) + sum(base_lens['G'])

        print (contig.name,
               length,
               args.min_run,
               count,
               max_len,
               min_len,
               tot_len,
               AT_len,
               CG_len,
               sum(base_lens['A']),
               sum(base_lens['T']),
               sum(base_lens['C']),
               sum(base_lens['G']),
               get_ratio(tot_len, length),
               get_ratio(AT_len, length),
               get_ratio(CG_len, length),
               get_ratio(sum(base_lens['A']), length),
               get_ratio(sum(base_lens['T']), length),
               get_ratio(sum(base_lens['C']), length),
               get_ratio(sum(base_lens['G']), length),
               sep='\t')


@command.add_sub
@argument('fasta', nargs='?', default='/dev/stdin')
@argument('--names', nargs='*', help='list of fasta names in order of fasta')
def fa_rename_all(args):
    """ Replace sequence names

    The number of names should be the same as that of sequences.
    """
    names = iter(args.names)

    for line in open(args.fasta):
        if line.startswith('>'):
            tokens = line.rstrip('\r\n').split(' ', 1)
            name = next(names)
            print ('>', name, ' ', ''.join(tokens[1:]), sep='')
        else:
            print (line, end='')  # avoiding doubled return

def _select_or_omit(fasta, names=None, omits=None):
    targets = names and set(names)
    un_targets = set(omits or [])

    for line in open(fasta):
        if line.startswith('>'):
            logging.debug('%s', line.rstrip())
            tokens = line.rstrip('\r\n').split(' ', 1)
            name = tokens[0][1:]
            rest = ''.join(tokens[1:])
            if name in un_targets:
                emit = False
            elif targets is None or name in targets:
                print ('>', name, ' ', rest, sep='')
                emit = True
            else:
                emit = False
        elif emit:
            print (line, end='')


@command.add_sub
@argument('fasta')
@argument.exclusive(
    argument('-n', '--names', nargs='*', help='list of fasta names to select'),
    argument('-N', '--name-file', help='list of fasta names to select'),
)
def fa_select(args):
    """ Select sequences by names

    O(n) but not required faidx
    """
    if args.name_file:
        names = [line.rstrip() for line in open(args.name_file)]
    else:
        names = args.names
    _select_or_omit(args.fasta, names=names)


@command.add_sub
@argument('fasta')
@argument.exclusive(
    argument('-n', '--names', nargs='*', help='list of fasta names to omit'),
    argument('-N', '--name-file', help='list of fasta names to omit'),
)
def fa_omit(args):
    """ Omit sequenses by names

    O(n) but not required faidx
    """
    if args.name_file:
        names = [line.rstrip() for line in open(args.name_file)]
    else:
        names = args.names
    _select_or_omit(args.fasta, omits=names)


@command.add_sub
@argument('fasta', nargs='?', default='/dev/stdin')
@argument('-f', '--from-name', required=True)
@argument('-t', '--to-name', required=True)
def fa_rename(args):
    """ Rename a sequence

    O(n) but not required faidx
    """
    from_name = args.from_name
    to_name = args.to_name

    for line in open(args.fasta):
        if line.startswith('>'):
            name = line.rstrip('\r\n').split(' ', 2)[0][1:]
            if name == from_name:
                print ('>', to_name, sep='')
            else:
                print ('>', name, sep='')
        else:
            print (line, end='')

@command.add_sub
@argument('fasta', nargs='?', default='/dev/stdin')
@argument('--prefix', default='')
@argument('--suffix', default='')
def fa_subname(args):
    """ Substitute fasta names
    """
    prefix = args.prefix
    suffix = args.suffix

    for line in open(args.fasta):
        if line.startswith('>'):
            name = line.rstrip('\r\n').split(' ', 2)[0][1:]
            name = prefix + name + suffix
            print ('>', name, sep='')
        else:
            print (line, end='')

@command.add_sub
@argument('fastq', nargs='?', default='/dev/stdin')
@argument('--prefix', default='')
@argument('--suffix', default='')
def fq_subname(args):
    """ Substitute fasta names
    """
    prefix = args.prefix
    suffix = args.suffix

    for i, line in enumerate(open(args.fastq)):
        if i % 4 == 0:
            name = line.rstrip('\r\n').split(' ', 2)[0][1:]
            name = prefix + name + suffix
            print ('@', name, sep='')
        else:
            print (line, end='')


@command.add_sub
@argument('fasta')
@argument('outdir')
def pb_fa_split(args):
    """ Split fasta to create pacbio input.fofn
    """
    logging.info('Making directory: %s', args.outdir)
    os.makedirs(args.outdir)  # this should be error on found
    run_lines = defaultdict(list)
    with io.open(args.fasta) as fp:
        for line in fp:
            line = line.rstrip('\r\n')
            if line.startswith('>'):
                run_name = line[1:].split('/')[0]
            run_lines[run_name].append(line)

    # TODO create pool of writers, or buffering
    for run_name in run_lines:
        path = os.path.join(args.outdir, '{0}.fa'.format(run_name))
        with io.open(path, '+a') as fp:
            for line in run_lines[run_name]:
                print (line, file=fp)


@command.add_sub
@argument('fastq', nargs='?', default='/dev/stdin')
@argument('--prefix', default='')
@argument('--suffix', default='')
@argument('--zmw-start', default=0)
def pb_fq_subname(args):
    """ Substitute fasta names
    """
    prefix = args.prefix
    suffix = args.suffix
    zmw = args.zmw_start
    with open(args.fastq) as fp:
        try:
            while 1:
                line1 = next(fp)
                name = line1.rstrip().split(' ', 2)[0][1:]
                seq = next(fp).rstrip()
                plus = next(fp).rstrip()
                qseq = next(fp).rstrip()
                length = len(seq)
                line1 = '@{}{}{}/{}/{}_{}'.format(prefix, name, suffix, zmw, 0, length)
                print (line1)
                print (seq)
                print (plus)
                print (qseq)
                zmw += 1
        except StopIteration:
            pass

@command.add_sub
@argument('fasta', nargs='?', default='/dev/stdin')
@argument('-u', '--uppercase', action='store_true')
@argument('-H', '--no-header', dest='header', action='store_false')
def fa2tsv(args):
    """ Substitute fasta names
    """
    sep = '\t'
    if args.header:
        print ('name', 'seq', sep=sep)
    with open(args.fasta) as fp:
        for contig in iter_contigs(fp):
            name = contig.name
            seq = contig.get_seq()
            if args.uppercase:
                seq = seq.upper()
            print (name, seq, sep=sep)


@command.add_sub
@argument('fasta', nargs='?', default='/dev/stdin')
@argument('-q', '--quality', default='I')
def fa2fq(args):
    """ Convert fasta to fastq
    """
    bq = args.quality
    class Fastq(object):
        __slots__ = ('name', '_seqs')
        def __init__(self, name=''):
            self.name = name
            self._seqs = []

        def append(self, seq):
            self._seqs.append(seq)

        def emit(self):
            seq = ''.join(self._seqs)
            print ('@' + self.name)
            print (seq)
            print ('+')
            print (bq * len(seq))

    with open(args.fasta) as fp:
        fastq = None
        try:
            while fp:
                line = next(fp).rstrip()
                if line.startswith('>'):
                    if fastq is not None:
                        fastq.emit()
                    name = line[1:]
                    fastq = Fastq(name)
                else:
                    fastq.append(line)
        except StopIteration:
            if fastq is not None:
                fastq.emit()


@command.add_sub
@argument('fastq', nargs='?', default='/dev/stdin')
def fq2fa(args):
    """ Convert fastq to fasta
    """
    with open(args.fastq) as fp:
        try:
            while fp:
                line = next(fp).rstrip()
                rname = line[1:]
                print ('>' + rname)
                seq = next(fp).rstrip() # seq
                print (seq)
                next(fp) # +
                next(fp) # quality
        except StopIteration:
            pass


@command.add_sub
@argument('fasta', nargs='?', default='/dev/stdin')
@argument('-n', '--name', required=True)
@argument('-e', '--edits', nargs='*', help='sequence of pos:alts')
@argument('--is-compact', action='store_true', default=False, help='input sequence is assumed compact ("-" is omitted)')
def fa_edit(args):
    """ Edit fasta sequence

    pos list (1-based) and alt bases
    ex: 123:AAA-T        (ref)
    ex: 123:AA-TT:AAA-T  ()
    """
    name = args.name
    contigs = list(iter_contigs(open(args.fasta)))
    logging.info('name: %s, edits: %s', name, args.edits)

    Edit = namedtuple('Edit', 'pos ref alt')
    def parse_edit(edit):
        els = edit.split(':')
        pos = int(els[0])
        if len(els) == 2:
            ref = None
            alt = els[1]
        elif len(els) == 3:
            ref = els[1]
            alt = els[2]
        else:
            raise Exception('Invalid edit {0}'.format(edit))
        if args.is_compact:
            alt = alt.replace('-', '')
            ref = ref and ref.replace('-', '')
        return Edit(pos, ref, alt)

    edits = list(sorted(map(parse_edit, args.edits), key=lambda e: e.pos))

    for contig in contigs:
        print ('>', contig.name_line, sep='')
        seq = contig.get_seq()
        if contig.name == name:
            bufs = []
            seq_pos = 1
            for edit in edits:
                bufs.append(seq[seq_pos - 1 : edit.pos - 1])  # prev seq
                bufs.append(edit.alt)
                if edit.ref is None:
                    ref_len = alt_len
                else:
                    ref_len = len(edit.ref)  # assume same length

                seq_pos = edit.pos + ref_len

            if seq_pos - 1 <= len(seq):
                bufs.append(seq[seq_pos - 1 :])

            len0 = len(seq)
            seq = ''.join(bufs)
            logging.info('seq length (%s -> %s)', len0, len(seq))
        print (seq)


_rev_maps = dict(zip('ATCGatcgNn', 'TAGCtagcNn'))

def _revcomp(seq):
    return ''.join(_rev_maps[c] for c in reversed(seq))


@command.add_sub
@argument('fasta', nargs='?', default='/dev/stdin', help='all contigs should be the same lengths')
@argument('-c', '--contig-names', nargs='*', default=None, help='list of reverse strand')
@argument('-a', '--add-suffix', action='store_true', help='add suffix to reversed')
def fa_reverse(args):
    """ Get reverse complements of sequences
    """
    names = set(args.contig_names)

    for contig in iter_contigs(args.fasta):
        reverse = contig.name in names

        name = contig.name
        if args.add_suffix:
            name = contig.name + ('.R' if reverse else '.F')

        print ('>' + name)
        seq = contig.seq
        print (_revcomp(seq) if reverse else seq)


@command.add_sub
@argument('regex')
@argument('fasta', nargs='?', default='/dev/stdin', help='all contigs should be the same lengths')
def fa_grep(args):
    """ Emit matched coordinates and sequence as BED format

    # TODO include context sequence
    """
    logging.info('pattern: %s', args.regex)
    pattern = re.compile(args.regex)
    for contig in iter_contigs(open(args.fasta)):
        seq = contig.get_seq().upper()
        for m in pattern.finditer(seq):
            print (contig.name, m.start(), m.end(), m.group(), sep='\t')



from __future__ import print_function
import os
from six import string_types
import logging
import multiprocess
import signal
import datetime
import time
from . import sh

class BaseQueue(object):
    def call(self, args, **kwds):
        raise NotImplementedError

    def array_call(self, args, **kwds):
        raise NotImplementedError


# deprecated
class ShellQueue(BaseQueue):
    def __init__(self, max_ntask=None, dry_run=False, **kwds):
        self.job_dir = None
        self.max_ntask = max_ntask
        self.dry_run = dry_run

    def _strip_ignores(self, kwds):
        ignores = ['name', 'hold', 'memory', 'queue', 'shell', 'script_name']
        for name in ignores:
            if kwds.pop(name, None):
                logging.info('Ignore option: {0}'.format(name))
        return kwds

    def call(self, args, slot=None, dry_run=None, **kwds):
        kwds = self._strip_ignores(kwds)
        if dry_run is None:
            dry_run = self.dry_run
        return sh.call(args, dry_run=dry_run, **kwds)

    def array_call(self, args, ntask, slot=None, max_ntask=None, dry_run=None, **kwds):
        kwds = self._strip_ignores(kwds)
        if dry_run is None:
            dry_run = self.dry_run
        assert isinstance(args, string_types)
        def init_pool():
            signal.signal(signal.SIGINT, signal.SIG_IGN)

        def task(n):
            task_id = n + 1
            env = dict(os.environ, TASK_ID=task_id)
            return self.call(args, dry_run=dry_run, env=env)

        try:
            pool = Pool(self.max_ntask, init_pool)
            for result in pool.imap_unordered(task, range(ntask)):
                pass
            pool.close()
        except Exception as e:
            logging.error(e)
            pool.terminate()
            raise e
        finally:
            pool.join()

LocalQueue = ShellQueue

# TODO load queue.ini

class UGEQueue(BaseQueue):
    """
    .job/{}
    """
    def __init__(self, max_ntask=None, job_dir='.job', dry_run=False):
        self.job_dir = job_dir
        self.dry_run = dry_run
        self.queue = ''
        self.max_ntask = max_ntask
        self.shell = '/bin/bash'
        self.script_name = 'job.sh'
        self.s_vmem_boost = 1.5
        self.slot = 1
        self.memory = 4.

    def _create_job_script(self, script_body, name=None, job_dir=None, script_name=None, shell=None, slot=None, memory=None, queue=None, hold=None, is_array=False, dry_run=None):
        if job_dir is None:
            job_dir = self.job_dir
        if name is None:
            name = 'job'
        if shell is None:
            shell = self.shell
        if script_name is None:
            script_name = self.script_name
        if dry_run is None:
            dry_run = self.dry_run
        queue = self.queue if queue is None else queue
        memory = self.memory if memory is None else memory
        slot = self.slot if slot is None else slot

        qopts = []
        # shell
        qopts.append(('-S', self.shell))

        # resource
        assert slot >= 1
        if slot > 1:
            qopts.append(('-pe', 'def_slot', str(slot)))
            memory = 1. * memory / slot
        qopts.append(('-l', '{queue},mem_req={mem_req}G,s_vmem={s_vmem}G'.format(
            queue=queue, mem_req=memory, s_vmem=memory * self.s_vmem_boost)))

        # job/log directory
        dt = datetime.datetime.now().strftime('%Y.%m.%d_%H.%M.%S.%f')
        data_dir = os.path.join(job_dir, '{name}_{dt}'.format(name=name, dt=dt))
        time.sleep(.00001)
        logging.info('Save job files to %s', data_dir)
        if not dry_run:
            os.makedirs(data_dir)

        if is_array:
            log_suffix = '.$TASK_ID.$JOB_ID.$HOSTNAME'
        else:
            log_suffix = '.$JOB_ID.$HOSTNAME'

        stdout_path = os.path.join(data_dir, 'stdout{0}'.format(log_suffix))
        stderr_path = os.path.join(data_dir, 'stderr{0}'.format(log_suffix))

        # other environment
        qopts.append(('-cwd',))
        qopts.append(('-V',))
        qopts.append(('-N', name))
        qopts.append(('-o', stdout_path))
        qopts.append(('-e', stderr_path))

        if hold:
            qopts.append(('-hold_jid', hold))

        lines = []
        lines.append('#!' + shell)
        for opt in qopts:
            opt = ' '.join(opt)
            lines.append('#$' + opt)

        lines.append(script_body)
        content = '\n'.join(lines)
        script_path = os.path.join(data_dir, script_name)
        logging.info('Script path: %s', script_path)
        logging.info('Stdout path: %s', stdout_path)
        logging.info('Stderr path: %s', stderr_path)
        logging.info('Script content:\n%s', content)
        if not dry_run:
            with open(script_path, 'w+') as writer:
                print (content, file=writer)
        return {'data_dir': data_dir, 'script_path': script_path}

    def _qsub_sync(self, script, ext_opts=None, dry_run=False):
        opts = ['-terse', '-sync', 'y']
        cmd = ['qsub'] + opts + list(ext_opts or []) + [script]
        sh.call(cmd, stdout='/dev/stderr', dry_run=dry_run)
        #from cStringIO import StringIO
        #out = StringIO()
        #sh.call(cmd, stdout=out)
        #if not dry_run:
        #    out.seek(0)
        #    job_id = out.read().rstrip()
        #    logging.info('%s (job_id: %s)', cmd, job_id)

    def _get_script_body(self, args):
        if isinstance(args, string_types):
            script_body = args
        else:
            script_body = ' '.join(map(sh.quote, args))
        return script_body

    def call(self, args, name=None, job_dir=None, slot=None, memory=None, hold=None, **kwds):
        script_body = self._get_script_body(args)
        dry_run = kwds.get('dry_run')
        info = self._create_job_script(script_body, name=name, hold=hold, job_dir=job_dir, slot=slot, memory=memory, dry_run=dry_run)
        script = info['script_path']
        return self._qsub_sync(script, dry_run=dry_run)

    def array_call(self, args, ntask, start=1, step=None, max_ntask=None, name=None, job_dir=None, slot=None, memory=None, hold=None, **kwds):
        script_body = self._get_script_body(args)

        assert ntask > 0, 'task number should be > 0'
        assert start > 0, 'start should be > 0'
        dry_run = kwds.get('dry_run')
        info = self._create_job_script(script_body, name=name, hold=hold, job_dir=job_dir, slot=slot, memory=memory, dry_run=dry_run, is_array=True)
        script = info['script_path']

        opts = []
        if step > 0:
            opts.extend(['-t', '{0}-{1}:{2}'.format(start, ntask, step)])
        else:
            opts.extend(['-t', '{0}-{1}'.format(start, ntask)])

        if max_ntask is not None:
            opts.extend(['-tc', max_ntask])

        return self._qsub_sync(script, dry_run=dry_run, ext_opts=opts)



from builtins import zip
import numpy as np


def get_banded_matrix(band1, band2, m, n, dtype=float):
    """
    >>> b = get_banded_matrix(1, 1, 2, 3, dtype=int)
    >>> b[1, 2] = 1
    >>> b[1, 2]
    1
    >>> b = get_banded_matrix(1, 1, 3, 2, dtype=int)
    >>> b[2, 1] = 1
    >>> b[2, 1]
    1
    >>> b[1, 2]
    0
    """
    assert band1 > 0 and band2 > 0
    data = np.zeros((band1 + band2 + 1, max(m, n)), dtype=dtype)
    return sparse_banded_matrix(data, -band1, m, n)


# similar interface to scipy.sparse.spdiags
class sparse_diag_matrix:
    """
    >>> data = np.array([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]])
    >>> diags = np.array([0, -1, 2])
    >>> sp = sparse_diag_matrix(data, diags, 4, 4)
    >>> sp.toarray()
    array([[1, 0, 3, 0],
           [1, 2, 0, 4],
           [0, 2, 3, 0],
           [0, 0, 3, 4]])
    >>> sp[0, 0]
    1
    >>> sp[2, 1]
    2
    >>> sp[3, 1]
    0
    """
    def __init__(self, data, diags, m, n):
        self._diags = dict((j1, vals) for j1, vals in zip(diags, data))
        self.shape = (m, n)

    def __getitem__(self, key):
        if isinstance(key, tuple):
            i1, i2 = key
            if i1 < 0 or i2 < 0:
                raise NotImplementedError('Cannot handle negative index')
            j1 = i2 - i1   # diag index
            j2 = i2
            d = self._diags.get(j1)
            return d[j2] if d is not None else 0
        raise NotImplementedError

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            i1, i2 = key
            if i1 < 0 or i2 < 0:
                raise NotImplementedError('Cannot handle negative index')
            j1 = i2 - i1   # diag index
            j2 = i2
            self._diags[j1][j2] = value   # cannot set values to non-initialized positions
            return
        raise NotImplementedError

    def toarray(self):  # adhoc
        m, n = self.shape
        from scipy.sparse import spdiags
        data = []
        diags = []
        for j1, dat in self._diags.items():
            data.append(dat)
            diags.append(j1)
        return spdiags(data, diags, m, n).toarray()


class sparse_banded_matrix:
    """
    >>> data = np.array([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]])
    >>> offset = -1
    >>> sp = sparse_banded_matrix(data, offset, 4, 4)
    >>> sp.toarray()
    array([[1, 2, 0, 0],
           [1, 2, 3, 0],
           [0, 2, 3, 4],
           [0, 0, 3, 4]])
    >>> sp[0, 0]
    1
    >>> sp[2, 1]
    2
    >>> sp[3, 1]
    0
    """
    def __init__(self, data, offset, m, n):
        self.shape = (m, n)
        self._offset = offset
        self._data = data
        assert data.shape[1] >= max(m, n), 'data shape should be longer than matrix length'
        self._width = data.shape[0]

    def __getitem__(self, key):   # FIXME not so fast
        if isinstance(key, tuple):
            i1, i2 = key
            if i1 < 0 or i2 < 0:
                raise NotImplementedError('Cannot handle negative index')
            j1 = i2 - i1 - self._offset   # diag index
            j2 = i2
            return self._data[j1, j2] if 0 <= j1 < self._width else 0
        raise NotImplementedError

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            i1, i2 = key
            if i1 < 0 or i2 < 0:
                raise NotImplementedError('Cannot handle negative index')
            j1 = i2 - i1 - self._offset  # diag index
            j2 = i2
            self._data[j1, j2] = value   # cannot set values to non-initialized positions
            return
        raise NotImplementedError

    def toarray(self):  # adhoc
        m, n = self.shape
        from scipy.sparse import spdiags
        data = self._data
        diags = list(range(self._offset, self._offset + self._width))
        return spdiags(data, diags, m, n).toarray()



from __future__ import print_function
from argtools import command, argument
import logging
import pysam


def iter_tabs(fp):
    for line in fp:
        if line.startswith('#'):
            continue
        yield line.rstrip('\r\n').split('\t')


@command.add_sub
@argument('region_bed')
@argument('bam')
@argument('--use-1start', action='store_true', default=False, help='regard region bed as 1 start and 1 end')
@argument('-m', '--mass', choices=['middle', 'start', 'end'], default='middle', help='mass point of read to determine the read is inside of the region')
def bam_coverage(args):
    """
    The following reads will be omitted
    - secondary alignment
    - supplementary alignment
    - unmapped

    Append following columns to input bed

    coverage of
    - all reads
    - pair-mapped
    - MQ0
    - MQ0 pair-mapped
    """
    offset = 1 if args.use_1start else 0

    get_mass = {
            'start': lambda rec: rec.pos,
            'end':   lambda rec: rec.aend - 1,   # end position of the alignment (note that aend points one past the last aligned residue)
            'middle': lambda rec: (rec.pos + rec.aend - 1) / 2.,   # middle position of the alignment
    }[args.mass]

    with pysam.Samfile(args.bam, 'rb') as samfile:
        for bed in iter_tabs(open(args.region_bed)):
            ref = bed[0]
            start = int(bed[1]) - offset
            end = int(bed[2])
            logging.info('fetching %s:%s-%s', ref, start, end)

            bam_recs = samfile.fetch(reference=ref, start=start, end=end)
            total = 0
            paired = 0
            total0 = 0
            paired0 = 0
            for rec in bam_recs:
                if rec.flag & 0x904:  # skip unmapped, secondary and supplementary alignment
                    continue
                mass = get_mass(rec)
                if not (start <= mass < end):  # skip if mass point is not included
                    continue

                is_paired = \
                    not rec.is_unmapped \
                    and not rec.mate_is_unmapped \
                    and rec.is_proper_pair

                total += 1
                if is_paired:
                    paired += 1

                if rec.mapq == 0:
                    total0 += 1
                    if is_paired:
                        paired0 += 1

            covs = (total, paired, total0, paired0)
            print (*(tuple(bed) + covs), sep='\t')



from __future__ import print_function, absolute_import
from argtools import command, argument
from collections import namedtuple
import os
from ..igv import IGVServer, IGVClient
import vcf


__SESSION_TMPL__ = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<Session {genome_attr} locus="{locus}" version="5">
    <Resources>
        {resources}
    </Resources>
    <Panel height="844" name="Panel1371449666544" width="{width}">
        {panel}
    </Panel>
    <Panel height="329" name="FeaturePanel" width="{width}">
        <Track altColor="0,0,178" autoScale="false" color="0,0,178" displayMode="COLLAPSED" featureVisibilityWindow="-1" fontSize="10" id="Reference sequence" name="Reference sequence" sortable="false" visible="true"/>
        <Track altColor="0,0,178" autoScale="false" clazz="org.broad.igv.track.FeatureTrack" color="0,0,178" colorScale="ContinuousColorScale;0.0;297.0;255,255,255;0,0,178" displayMode="COLLAPSED" featureVisibilityWindow="-1" fontSize="10" height="35" id="hg19_genes" name="RefSeq Genes" renderer="BASIC_FEATURE" sortable="false" visible="true" windowFunction="count">
            <DataRange baseline="0.0" drawBaseline="true" flipAxis="false" maximum="297.0" minimum="0.0" type="LINEAR"/>
        </Track>
    </Panel>
    <PanelLayout dividerFractions="{dividerFractions}"/>
</Session>"""


@command.add_sub
@argument('-g', '--genome')
@argument('-l', '--locus', default='All')
@argument('files', nargs='*')
def igv_session(args):
    """
    """
    paths = []
    path_set = set()
    tracks = []
    for path in args.files:
        path = os.path.abspath(path)
        if path not in path_set:
            paths.append(path)
        if path.endswith('.bam'):
            name = os.path.basename(path)
            track = """\
<Track altColor="0,0,178" autoScale="true" color="175,175,175" displayMode="{displayMode}" featureVisibilityWindow="-1" fontSize="10" id="{id}" name="{name}" showReference="false" snpThreshold="{snpThreshold}" sortable="true" visible="true">
    <DataRange baseline="0.0" drawBaseline="true" flipAxis="false" maximum="60.0" minimum="0.0" type="LINEAR"/>
</Track>""".format(
                id=path + '_coverage',
                name=name + ' Coverage',
                displayMode='COLLAPSED',
                snpThreshold=0.2,
            )
            tracks.append(track)

        # default
        name = os.path.basename(path)
        track = """\
<Track altColor="0,0,178" autoScale="false" color="0,0,178" displayMode="{displayMode}" featureVisibilityWindow="-1" fontSize="10" id="{id}" name="{name}" showSpliceJunctions="false" sortable="true" visible="true">
        <RenderOptions colorByTag="" colorOption="{colorOption}" flagUnmappedPairs="false" groupByTag="" maxInsertSize="1000" minInsertSize="50" shadeBasesOption="{shadeBasesOption}" shadeCenters="true" showAllBases="false" sortByTag=""/>
</Track>""".format(
            id=path,
            name=name,
            displayMode='EXPANDED',
            colorOption='UNEXPECTED_PAIR',
            shadeBasesOption='QUALITY',
        )
        tracks.append(track)

    genome = args.genome
    if genome.endswith('.fa'):
        genome = os.path.abspath(genome)

    session = __SESSION_TMPL__.format(
            resources='\n'.join('<Resource path="{path}"></Resource>'.format(path=path) for path in paths),
            panel='\n'.join(tracks),
            genome_attr='genome="{0}"'.format(genome) if genome else '',
            locus=args.locus,
            width=2528,
            dividerFractions='0,0.7',  # TODO
            #dividerFractions='0.005046257359125316,0.7190916736753574',  # TODO
    )
    print (session)


@command.add_sub
@argument('feature', help='default is bed file')
@argument('-t', '--type', default=None, help='feature type')
@argument('-s', '--session')
@argument('-p', '--igv-port', default=60151, type=int)
@argument('-r', '--margin-ratio', default=.1, type=float)
@argument('-m', '--min-length', default=300, type=int)
def igv_links(args):
    """ Output html to jump region
    You may add session loading link by specify <session>

    BED
    BED with header
    VCF
    """
    session = os.path.abspath(args.session)
    Region = namedtuple('Region', 'contig start end')
    def iter_vcf(fname):
        with open(fname) as fp:
            reader = vcf.Reader(fp)
            header = ['contig', 'start', 'end', 'length', 'id', 'ref', 'alt', 'filter', 'info']
            yield header
            for rec in reader:
                start = rec.POS - 1
                end = rec.INFO['END'] if 'END' in rec.INFO \
                        else rec.POS + len(str(rec.REF)) - 1
                reg = Region(rec.CHROM, start, end)
                info = ' '.join('{0}={1}'.format(k, v) for k, v in rec.INFO.items())
                alt = ','.join(map(str, rec.ALT))
                row = [rec.CHROM, start, end, end - start, rec.ID, rec.REF, alt, ''.join(rec.FILTER), info]
                yield reg, row

    def iter_bed(fname):  # TODO how to handle with header
        with open(fname) as fp:
            row = next(fp).split('\t')
            if row[1].isdigit():
                # no header
                header = ['contig', 'start', 'end', 'length', 'name'] + list('col{0}'.format(i+1) for i in xrange(len(row) - 4))
                reg = Region(row[0], int(row[1]), int(row[2]))
                row = row[:2] + [reg.end - reg.start] + row[2:]
                yield header
                yield reg, row  # first line
            else:
                yield row  # header
            for row in fp:
                reg = Region(row[0], int(row[1]), int(row[2]))
                row = row[:2] + [reg.end - reg.start] + row[2:]
                yield reg, row

    its = {'bed': iter_bed, 'vcf': iter_vcf}  # TODO bed with head
    def get_iterable(fname):
        if args.type is not None:
            return its[args.type](fname)
        if fname.endswith('.vcf'):
            return its['vcf'](fname)
        if fname.endswith('.bed'):
            return its['bed'](fname)
        return its['bed'](fname)

    print ('''<html>''')
    print ('''<body>''')
    print ('''<p><a href="http://localhost:{igv_port}/load?file=file://{session}"> Load session </a></p>'''.format(igv_port=args.igv_port, session=session))
    print ('<table>')
    it = get_iterable(args.feature)
    header = next(it)
    print ('<thead>')
    print ('<tr>')
    print ('''<th>IGV link</th>''')
    for term in header:
        print ('<th>{0}</th>'.format(term))
    print ('</tr>')
    print ('</thead>')
    print ('<tbody>')
    for reg, row in it:
        length = reg.end - reg.start
        margin = int(length * args.margin_ratio)
        left = reg.start - margin
        right =  reg.end + margin
        diff = args.min_length - (right - left)
        if diff > 0:
            left -= int(diff / 2)
            right += int(diff / 2)
        locus = '{0}:{1}-{2}'.format(reg.contig, left, right)
        print ('<tr>')
        print ('''<td><a href="{link}">igv_link</a></td>'''.format(link='http://localhost:{igv_port}/goto?locus={locus}'.format(igv_port=args.igv_port, locus=locus)))
        for term in row:  # TODO html escape
            print ('<td>{0}</td>'.format(term))
        print ('</tr>')
    print ('</tbody>')
    print ('</table>')
    print ('''</body>''')
    print ('''</html>''')


@command.add_sub
@argument('files', nargs='+', help='URL or file path')
@argument.exclusive(
    argument('-l', '--loci', nargs='+'),
    argument('-L', '--loci-file'),
)
@argument('-P', '--paths', nargs='+')
@argument('-g', '--genome')
@argument('-p', '--port', type=int)
@argument('-o', '--outdir')
@argument('-m', '--view-mode', choices=['expand', 'collapse', 'squished'])
@argument('--view-as-pairs', action='store_true')
@argument('--height', type=int)
@argument('--run-server', action='store_true', help='Run IGV server during snapshot')
@argument('--no-reset', dest='reset_on_end', action='store_false', help='Reset IGV server status on end')
def igv_snapshot(args):
    """ Taking IGV snapshot
    """
    if not args.loci and not args.loci_file:
        logging.warning('Either of loci or loci_file is required')
        return (1)

    def iter_loci():
        if args.loci:
            for locus in args.loci:
                locus_name = locus.replace(':', '_').replace('/', '_') + '.png'
                path = os.path.join(args.outdir or os.curdir, locus_name)
                yield (locus, path)
        elif args.loci_file:
            with open(args.loci_file) as fp:
                for line in fp:
                    row = line.rstrip().split('\t')
                    locus = row[0]
                    locus_name = locus.replace(':', '_').replace('/', '_') + '.png'
                    path = os.path.join(args.outdir or os.curdir, locus_name)
                    yield (locus, path)

    def snapshot(igv):
        if args.genome:
            igv.genome(args.genome)
        if args.height:
            igv.set_height(args.height)
        for fn in args.files:
            igv.load(fn)
        if args.view_mode:
            getattr(igv, view_mode)()
        if args.view_as_pairs:
            igv.view_as_pairs()

        for locus, path in iter_loci():
            igv.goto(locus)
            igv.save(path)
        if args.reset_on_end:
            igv.new()

    if args.run_server:
        with IGVServer(port=args.port) as server, server.get_client() as igv:
            snapshot(igv)
    else:
        with IGVClient(port=args.port) as igv:
            snapshot(igv)


if __name__ == '__main__':
    command.run()



from __future__ import print_function
import logging
import numpy as np
from ..matrix import get_banded_matrix
from .basics import *


#TODO shadow class for doctest

class _NWAlignmentTrace(basics.NWAlignmentTrace):
    """
    >>> score = LinearGapScore(1, -2, -2, -2)
    >>> seq1 = 'ATATGCTATAT'
    >>> seq2 = 'ATATCTGATCT'
    >>> a = NWAlignmentTrace(seq1, seq2, move_score=score)
    >>> r = a.get_best()
    >>> r['aln1']
    'ATATGCT-ATAT'
    >>> r['aln2']
    'ATAT-CTGATCT'
    >>> r['score'] == 4 - 2 + 2 - 2 + 2 - 2 + 1
    True
    """


class _NWAlignment(NWAlignment):
    """
    >>> score = LinearGapScore(1, -2, -2, -2)
    >>> seq1 = 'ATATGCTATAT'
    >>> seq2 = 'ATATCTGATCT'
    >>> a = NWAlignment(seq1, seq2, move_score=score)
    >>> r = a.get_best()
    >>> r['aln1']
    'ATATGCT-ATAT'
    >>> r['aln2']
    'ATAT-CTGATCT'
    >>> r['score'] == 4 - 2 + 2 - 2 + 2 - 2 + 1
    True
    """


class _SWAlignment(SWAlignment):
    """
    >>> score = LinearGapScore(1, -2, -2, -2)
    >>> aln1 = 'ATATGCTGATAT'
    >>> aln2 = 'ATAT-CTGATCT'
    >>> seq1 = aln1.replace('-', '')
    >>> seq2 = aln2.replace('-', '')
    >>> a = SWAlignment(seq1, seq2, move_score=score)
    >>> r = a.get_best()
    >>> r['start1'], r['aln1']
    (1, 'ATATGCTGAT')
    >>> r['start2'], r['aln2']
    (1, 'ATAT-CTGAT')
    >>> r['score'] == 4 - 2 + 5 == 7
    True

    >>> aln1 = 'AAGATATGCTAATAT'
    >>> aln2 = '----TAT-CTAA---'
    >>> seq1 = aln1.replace('-', '')
    >>> seq2 = aln2.replace('-', '')
    >>> a = SWAlignment(seq1, seq2, move_score=score)
    >>> r = a.get_best()
    >>> r['start1'], r['aln1']
    (5, 'TATGCTAA')
    >>> r['start2'], r['aln2']
    (1, 'TAT-CTAA')
    >>> r['score'] == 3 - 2 + 4 == 5
    True
    """


class _NWAlignmentBandedNaive(NWAlignmentBandedNaive):
    """
    >>> score = LinearGapScore(1, -2, -2, -2)
    >>> seq1 = 'ATATGCTATAT'
    >>> seq2 = 'ATATCTGATCT'
    >>> a = NWAlignmentBandedNaive(seq1, seq2, band1=3, band2=3, move_score=score)
    >>> r = a.get_best()
    >>> r['aln1']
    'ATATGCT-ATAT'
    >>> r['aln2']
    'ATAT-CTGATCT'
    >>> r['score'] == 4 - 2 + 2 - 2 + 2 - 2 + 1
    True
    """


class _NWAlignmentBandedSparse(NWAlignmentBandedNaive):
    """
    >>> score = LinearGapScore(1, -2, -2, -2)
    >>> seq1 = 'ATATGCTATAT'
    >>> seq2 = 'ATATCTGATCT'
    >>> a = NWAlignmentBandedSparse(seq1, seq2, band1=3, band2=3, move_score=score)
    >>> r = a.get_best()
    >>> r['aln1']
    'ATATGCT-ATAT'
    >>> r['aln2']
    'ATAT-CTGATCT'
    >>> r['score'] == 4 - 2 + 2 - 2 + 2 - 2 + 1
    True
    """

#NWAlignmentBanded = NWAlignmentBandedSparse
#NWAlignmentBanded = NWAlignmentBandedNaive

class _NWGAlignment(NWGAlignment):
    """
    >>> score = AffineGapScore(match=1, mismatch=-2, gap1_open=-2, gap1_ext=-1, gap2_open=-2, gap2_ext=-1)
    >>> aln1 = 'AT--G'
    >>> aln2 = 'ATAAG'
    >>> seq1 = aln1.replace('-', '')
    >>> seq2 = aln2.replace('-', '')
    >>> a = NWGAlignment(seq1, seq2, move_score=score)
    >>> r = a.get_best()
    >>> (r['aln1'], r['aln2']) == (aln1, aln2)
    True
    >>> r['score'] == 2 - 2 - 1 + 1 == 0
    True

    >>> aln1 = 'ATATATATATATGGG--ATAT'
    >>> aln2 = '--------ATATGGGCCATCT'
    >>> seq1 = aln1.replace('-', '')
    >>> seq2 = aln2.replace('-', '')
    >>> a = NWGAlignment(seq1, seq2, move_score=score)
    >>> r = a.get_best()
    >>> (r['aln1'], r['aln2']) == (aln1, aln2)
    True
    >>> r['score'] == -8 + 7 - 2 - 1 + 2 - 2 + 1 == -3
    True

    >>> aln1 = 'ATATGCTGATAT-----'
    >>> aln2 = '-----CTGATCTNNNNG'
    >>> seq1 = aln1.replace('-', '')
    >>> seq2 = aln2.replace('-', '')
    >>> a = NWGAlignment(seq1, seq2, move_score=score)
    >>> r = a.get_best()
    >>> (r['aln1'], r['aln2']) == (aln1, aln2)
    True
    >>> r['score'] == -5 + 5 - 2 + 1 - 2 - 4 == -7
    True
    """



import os
from argtools import command, argument
from .. import sh
from ..utils import iter_tabs
import numpy as np


def random_rgb():
    return (np.random.random_integers(256),
            np.random.random_integers(256),
            np.random.random_integers(256))


#@command.add_sub
@argument('segdup_tab')
def segdup2trackbed(args):
    """
    * set unique name to pairs (use uid)
    * randomize color by name
    """
    with open(args.segdup_tab) as fp:
        rows = iter_tabs(fp)
        header = with_header(rows)
        for row in rows:
            chrom = row['chrom']
            start = row['chromStart']
            end = row['chromEnd']
            name = row['name']
            score = row['score']
            strand = row['strand']
            other_chrom = row['otherChrom']
            other_start = row['otherStart']
            other_end = row['otherEnd']


def require(file):
    if not os.path.exists(file):
        raise Exception('File does not exist: {0}'.format(file))


@command.add_sub
@argument('region_bed', help='with header')
@argument('fasta', help='reference fasta')
@argument('-db', required=True, help='blast db')
@argument('--col-id', default='name', help='column name of region id')
@argument('-o', '--out-dir', default='refdup_analyze')
@argument('-j', '--njobs', type=int, default=8)
@argument('-f', '--force-rerun', action='store_true')
@argument('-m', '--min-size', type=int, default=500)
def bed_refdup_analyze(args):
    """
    Outputs:
        {outdir}/region.fa
        {outdir}/region.ref.blastn
        {outdir}/region.summary.txt
    """
    out_dir = args.out_dir
    region_bed = '{0}/region.bed'.format(out_dir)
    region_fasta = '{0}/region.fa'.format(out_dir)
    region_blastn = '{0}/region.blastn'.format(out_dir)
    region_summary = '{0}/region.summary.txt'.format(out_dir)

    sh.call(['mkdir', '-p', out_dir])

    # create copy of input bed
    if args.force_rerun or not os.path.exists(region_bed):
        sh.call('cat {0} | awk "NR==1; NR>1 && $3-$2 > {min_size}"'.format(
            args.region_bed, min_size=args.min_size),
            stdout=region_bed)

    # make region fasta
    if args.force_rerun or not os.path.exists(region_fasta):
        require(region_bed)
        require(args.fasta)
        sh.call("""
        bedtools getfasta -name -bed <(csvcut -t -c chrom,start,end,{id} {bed} | csvformat -T | sed 1d) -fi {fasta} -fo {out}
        """.format(id=args.col_id, bed=region_bed, fasta=args.fasta, out=region_fasta))

    # make region blastn
    if args.force_rerun or not os.path.exists(region_blastn):
        require(region_fasta)
        sh.call("""blastn -db {db} -num_threads {njobs} -query {fasta} -evalue 0.1 -outfmt '7 std gaps qlen slen qseq sseq'"""
            .format(db=args.db, fasta=region_fasta, njobs=args.njobs),
            stdout=region_blastn)




from __future__ import print_function
from argtools import command, argument
from builtins import zip
import logging
import pandas as pd
import gzip
import tempfile
import glob
from collections import namedtuple
from ..utils import file_split, iter_tabs, collect_while
from .. import sh


@command.add_sub
@argument('input')
@argument('--sep', default='\t')
def transpose(args):
    with open(args.input) as fp:
        for c in zip(*(l.strip().split(args.sep) for l in fp.readlines() if l.strip())):
             print (*c, sep=args.sep)


@command.add_sub   # similar to 'csvcut -t -n table'
@argument('input')
def header(args):
    with open(args.input) as fp:
        header = next(iter_tabs(fp), None)
        if header:
            for name in header:
                print (name)


@command.add_sub   # similar to 'csvcut -t -n table'
@argument('value')
@argument('-i', '--input', default='/dev/stdin')
@argument('-n', '--name', help='column name to add')
@argument('-c', '--column', type=int, help='Where to add column. default is last')
# argument('--no-template')
# argument('--eval')
# argument('--no-header')
def add_column(args):
    """
    add_column 'column_{col_name or col_number}' -n name
    add_column 'column_{__index__:05d}' -n id
    """
    use_header = bool(args.name)

    if args.column is None:  # last
        add_col = lambda row, value: row.append(value)
    else:
        add_col = lambda row, value: row.insert(args.column, value)

    with open(args.input) as fp:
        it = iter_tabs(fp)
        if use_header:
            header = next(it)
            add_col(header, args.name)
            print (*header, sep='\t')

        kwds = {}
        for __index__, row in enumerate(it, 1):
            if use_header:
                kwds = dict(zip(header, row))
            value = args.value.format(__index__=__index__, **kwds)
            add_col(row, value)
            print (*row, sep='\t')


@command.add_sub
@argument('file1')
@argument('file2')
def set_isec(args):
    """
    * tsv file
    * key is leftmost column

    # TODO impl. for many files
    """

    key_idx = 0
    with open(args.file1) as fp:
        keys1 = [row[key_idx] for row in iter_tabs(fp)]
    with open(args.file2) as fp:
        keys2 = [row[key_idx] for row in iter_tabs(fp)]

    key_set1 = set(keys1)
    key_set2 = set(keys2)
    key_isec = key_set1 & key_set2

    for key in keys1:
        if key in key_isec:
            print (key)


@command.add_sub
@argument('input', nargs='?', default='/dev/stdin')
@argument('-c', '--columns')
@argument('-C', '--not-columns')
def cut(args):
    with open(args.input) as fp:
        it = iter_tabs(fp)
        header = next(it, None)
        if header is None:
            return
        col_idxs = dict((col, i) for i, col in enumerate(header))
        if args.columns:
            columns = args.columns.split(',')
            idxs = [col_idxs[col] for col in columns]
        else:
            idxs = list(range(len(header)))

        if args.not_columns:
            not_idxs = set([col_idxs[col] for col in args.not_columns.split(',')])
            idxs = [i for i in idxs if i not in not_idxs]

        print (*(header[i] for i in idxs), sep='\t')
        for row in it:
            print (*(row[i] for i in idxs), sep='\t')


@command.add_sub
@argument('inputs', nargs='*', default='/dev/stdin')
@argument('--header', action='store_true')
def cat(args):
    """ concat inputs with header
    """
    for i, fn in enumerate(args.inputs):
        if i == 0:
            sh.call(['cat', fn])
        else:
            if args.header:
                sh.call(['sed', '1d', fn])
            else:
                sh.call(['cat', fn])


# TODO .gz
@command.add_sub
@argument('input')
@argument('-n', '--nlines', default=1000, type=int)
@argument('-p', '--prefix', default='')
@argument('-s', '--suffix', default='')
@argument('-d', '--dir', default=None)
@argument('--header', type=int, default=0, help='number of header lines')
def split(args):
    logging.info('The number of lines per file: %s', args.nlines)
    assert args.header >= 0
    with open(args.input) as fp:
        header = '\n'.join([next(fp).rstrip() for _ in range(args.header)])

        for name in file_split(fp, prefix=args.prefix, suffix=args.suffix, dir=args.dir, nlines=args.nlines, header=header):
            logging.info('Write to %s', name)
            print (name)


class _group_fns:
    @staticmethod
    def first(vals):
        return vals[0]

    @staticmethod
    def concat(vals, sep=','):
        return sep.join(vals)

    @staticmethod
    def length(vals):
        return len(vals)

    @staticmethod
    def sum(vals):
        return sum(map(float, vals))

_group_fn_list = [_attr for _attr in dir(_group_fns)
        if (not _attr.startswith('__') and not _attr.endswith('__') and callable(getattr(_group_fns, _attr)))]

@command.add_sub
@argument('-c', '--columns', nargs='+', help='column specs')
@argument('--no-header', '-H', dest='use_header', action='store_false')
@argument('--default-fn', '-f', default='first', choices=_group_fn_list, help='default group function')
@argument.exclusive(
argument('-i', '--inputs', nargs='+', help='input files'),
argument('-l', '--input-list', help='list of input file names'),
required=True
)
def summarize_row(args):
    """
    -c {col1} {col2} ...
    Availble specifiers :
        col1:col2               ---  concat(col1)
        col1::first
        col1:col1-sum:sum
        col1::concat            ---  concat(col1) as col1
        col1::length            ---  length(col1) as col1
        0                       ---  default_fn(filelist) (1st column) as filelist
        +n:sample               ---  (n+1)-th column of filelist
    """
    if not args.use_header:
        raise NotImplementedError('Not implemented yet')

    input_tab = None
    if args.inputs:
        logging.info('Reading %s files', len(args.inputs))
        fnames = list(args.inputs)
    elif args.input_list:
        logging.info('Read from %s', args.input_list)
        input_tab = pd.read_table(args.input_list, header=None)
        fnames = list(input_tab[0])
    else:
        logging.error('Require input file list')
        return 1

    rows_list = []  # [rows]
    headers = []  # [header]
    # TODO read using threads ?
    for i, fname in enumerate(fnames):
        logging.info('Loading file %s', fname)
        if fname.endswith('.gz'):
            _open = gzip.open
        else:
            _open = open
        with _open(fname) as fp:
            it = iter_tabs(fp)
            header = next(it)
            headers.append(header)
            if i == 0:
                for row in it:
                    rows_list.append([row])
            else:
                for j, row in enumerate(it):
                    rows_list[j].append(row)

    col_specs = []  # (col_name, getter, fn_name)
    if args.columns:
        columns = args.columns
    else:
        columns = headers[0]  # use header of first sample

    for col_spec in columns:
        col_specs.append(parse_col_spec(col_spec, headers=headers, fnames=fnames, input_tab=input_tab, default_fn=args.default_fn))

    print (*[sp.name for sp in col_specs], sep='\t')
    for rows in rows_list:
        out_row = []
        for sp in col_specs:
            vals = [sp.getter(row, fidx) for fidx, row in enumerate(rows)]
            out_val = sp.fn(vals)
            out_row.append(out_val)
        print (*out_row, sep='\t')


Colspec = namedtuple('ColSpec', 'name getter fn')
def parse_col_spec(col_spec, headers, fnames, input_tab, default_fn='first'):
    tokens = col_spec.split(':')
    if len(tokens) == 3:
        col, col_name, fn_name = tokens
        if not col_name:
            col_name = col
    elif len(tokens) == 2:
        col, col_name = tokens
        fn_name = None
    elif len(tokens) == 1:
        col, = tokens
        col_name = col
        fn_name = None
    else:
        raise NotImplementedError('Unreadable col_spec: {0}'.format(col_spec))

    if col == '0':
        getter = lambda row, fidx: fnames[fidx]
    elif col[0] == '+' and col[1:].isdigit():
        col_idx = int(col[1:])
        getter = lambda row, fidx: input_tab[col_idx][fidx]
    else:
        col_idxs = [header.index(col) for header in headers]
        getter = lambda row, fidx: row[col_idxs[fidx]]

    if fn_name is None:
        fn_name = default_fn
    fn = getattr(_group_fns, fn_name)

    return Colspec(name=col_name, getter=getter, fn=fn)


@command.add_sub
@argument('-c', '--columns', nargs='+', help='column specs')
@argument('--no-header', '-H', dest='use_header', action='store_false')
@argument('--default-fn', '-f', default='first', choices=_group_fn_list, help='default group operator')
@argument('-l', '--input-list', help='list of input file names', required=True)
@argument('-s', '--chunk-size', default=5000, help='default operator')
@argument('--dir', help='working directory')
@argument('--memory-reduce', type=float, default=8.)
@argument('--read-cmd', default='cat {0}', help='read cmd for input list')
#@argument('--keep-dir', action='store_true')
def summarize_row_chunked(args):
    """
    """
    if not args.use_header:
        raise NotImplementedError('Not implemented yet')

    from ..job_queue import UGEQueue
    queue = UGEQueue()

    if args.dir:
        work_dir = args.dir
        sh.call('mkdir -p {dir}'.format(dir=work_dir))
    else:
        work_dir = tempfile.mkdtemp()
    logging.info('Working directory: %s', work_dir)

    logging.info('Read from %s', args.input_list)
    input_tab = pd.read_table(args.input_list, header=None)
    fnames = list(input_tab[0])
    input_len = len(fnames)
    input_fn = '{dir}/input.fn'.format(dir=work_dir)
    input_tab[[0]].to_csv(input_fn, header=False, index=False, sep='\t')  # save only file names

    import sys
    galgo = ' '.join([sys.executable, sys.argv[0]])

    # create dirs
    sh.call('mkdir -p {dir}/inputs'.format(dir=work_dir))
    sh.call('mkdir -p {dir}/splits'.format(dir=work_dir))

    # split
    # less is used for read gz file if possible
    read_cmd = args.read_cmd.format('$fname')
    cmd = '''
    mkdir -p {dir}/inputs/$SGE_TASK_ID
    fname="$(sed -n ${{SGE_TASK_ID}}p {input_fn})"
    {galgo} split --dir {dir}/inputs/$SGE_TASK_ID --header 1 -n {nlines} --suffix .txt <({read_cmd})
    '''.format(galgo=galgo, dir=work_dir, nlines=args.chunk_size, input_fn=input_fn, read_cmd=read_cmd)
    queue.array_call(cmd, name='summarize_row_split', memory=8, ntask=input_len)

    first_splits = glob.glob('{dir}/inputs/1/*.txt'.format(dir=work_dir))
    split_len = len(first_splits)
    logging.info('Split to %s files.', split_len)

    # create input_list for each split job
    for split_idx in range(1, split_len + 1):
        input_files = [
            '{dir}/inputs/{input_idx}/{split_idx}.txt'.format(dir=work_dir, input_idx=input_idx, split_idx=split_idx)
            for input_idx in range(1, input_len + 1)
        ]
        input_list1 = '{dir}/splits/{split_idx}.list'.format(dir=work_dir, split_idx=split_idx)
        input_tab1 = input_tab.copy()
        input_tab1[0] = input_files  # overwrite with split files
        logging.info('Saving list %s', input_list1)
        input_tab1.to_csv(input_list1, header=False, index=False, sep='\t')

    # setup summarize_row options
    summary_opts = []
    if not args.use_header:
        summary_opts.append('-H')
    if args.default_fn:
        summary_opts.append('-f {0}'.format(args.default_fn))
    if args.columns:
        summary_opts.append('-c {0}'.format(' '.join(args.columns)))  #TODO quote

    cmd = '''
    {galgo} summarize_row -v -l {dir}/splits/$SGE_TASK_ID.list {opts} | gzip -c > {dir}/splits/$SGE_TASK_ID.txt.gz
    '''.format(galgo=galgo, dir=work_dir, opts=' '.join(summary_opts))
    queue.array_call(cmd, name='summarize_row_reduce', memory=args.memory_reduce, ntask=split_len)

    # concat
    logging.info('concat all')
    for split_idx in range(1, split_len + 1):
        if split_idx == 1:
            sh.call('zcat {dir}/splits/{split_idx}.txt.gz'.format(dir=work_dir, split_idx=split_idx))
        else:
            sh.call('zcat {dir}/splits/{split_idx}.txt.gz | sed 1d'.format(dir=work_dir, split_idx=split_idx))


@command.add_sub
@argument('input_xsls', help='input.xsls')
@argument('-l', '--list', action='store_true', help='list sheetnames')
@argument('-n', '--sheet-name', default='0')
@argument('-s', '--sep', default='\t')
@argument('-o', '--output', default='/dev/stdout')
def excel(args):
    xls = pd.ExcelFile(args.input_xsls)
    if args.list:
        for name in xls.sheet_names:
            print (name)
        return 0

    n = args.sheet_name
    if args.sheet_name.isdigit():
        n = int(args.sheet_name)
    parsed = xls.parse(sheetname=n)
    parsed.to_csv(args.output, header=True, index=False, sep=args.sep)


@command.add_sub
@argument('-o', '--output', default='output.xls', help='output.xsls')
@argument('-i', '--inputs', required=True, nargs='+', default='tsv files')
@argument('-n', '--names', nargs='+')
def excelify(args):
    logging.info('Writing to %s', args.output)
    writer = pd.ExcelWriter(args.output, engine='xlsxwriter')
    names = args.names
    for i, ifile in enumerate(args.inputs):
        tab = pd.read_table(ifile, dtype='str')
        if names:
            sheet_name = names[i]
        else:
            sheet_name = 'Sheet{0}'.format(i + 1)
        logging.info('Writing %s', sheet_name)
        tab.to_excel(writer, sheet_name, index=False)
    writer.save()


@command.add_sub
@argument('table')
@argument('-c', '--column')
@argument('-o', '--output', default='plot_hist.pdf')
@argument('-b', '--bins', default=50, type=int)
@argument('--xlim', nargs=2, type=float)
@argument('--ylim', nargs=2, type=float)
def plot_hist(args):
    logging.info('Save plot to %s', args.output)
    tab = pd.read_table(args.table)
    print (tab.describe())
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    plt.figure()
    ax = plt.gca()
    if args.xlim:
        ax.set_xlim(args.xlim[0], args.xlim[1])
    if args.ylim:
        ax.set_ylim(args.ylim[0], args.ylim[1])
    tab[args.column].hist(ax=ax, bins=args.bins)
    plt.savefig(args.output)
    plt.close()


@command.add_sub
@argument('pstat_file')
@argument('-s', '--sort', choices=['time', 'calls', 'cumulative', 'stdname'], default='time')
def show_pstats(args):
    """ Utility for parse cProfile result
    """
    import pstats
    st = pstats.Stats(args.pstat_file)
    st.sort_stats('time').print_stats()  # stdname calls time cumulative


@command.add_sub
@argument('-m', '--memory', default=4, type=float, help='Requesting memory in gigabyte')
#@argument('--smem-boost', default=1.5, type=float, help='difference of smem')
@argument('-q', '--queue', default=None)
@argument('-N', '--name', default=None)
@argument('-t', '--task', default=None)
#@argument('--hold_jid')
@argument('--hold', default=None)
@argument('-s', '--slot', default=1, type=int)
@argument('--dry-run', action='store_true')
@argument('-e', '--engine', choices=['uge', 'local'], default='local')
@argument('args', nargs='+')
def run(args):
    from ..job_queue import UGEQueue, LocalQueue
    if args.engine == 'local':
        queue = LocalQueue()
    elif args.engine == 'uge':
        queue = UGEQueue()
    else:
        raise NotImplementedError(args.engine)

    if len(args.args) == 1:
        cmd = args.args[0]
    else:
        cmd = args.args

    def parse_task(task):
        # 'n[-m][:s]'
        sp = task.split(':')
        if len(sp) >= 2:
            rg, step = sp[0], int(sp[1])
        else:
            rg, step = sp[0], None
        sp = rg.split('-')
        if len(sp) >= 2:
            start, ntask = int(sp[0]), int(sp[1])
        else:
            start = ntask = int(sp[0])
        return start, ntask, step

    if args.task:
        start, ntask, step = parse_task(args.task)
        queue.array_call(cmd, queue=args.queue, memory=args.memory, name=args.name, hold=args.hold, 
                start=start, ntask=ntask, step=step, slot=args.slot, dry_run=args.dry_run)
    else:
        queue.call(cmd, queue=args.queue, memory=args.memory, name=args.name, hold=args.hold, slot=args.slot, dry_run=args.dry_run)




from __future__ import print_function
from argtools import command, argument
from pysam import Samfile


@command.add_sub
@argument('vcf')
@argument('bam')
@argument('--fasta')
def bam_variant_aln(args):
    samfile = Samfile(args.bam)
    for rec in args.vcf:
        fp = open(vcf)
        reader = pyvcf.Reader(fp)
        self.positions = []

    for rec in samfile.fetch(vcf):
        samfile.getrname(rec.tid)
        rec



from builtins import zip
from collections import namedtuple
from operator import attrgetter
import logging
from .utils import color_term, cached_property, fill_text, chunked, iter_counts
import re
import numpy as np

_dna_rev_maps = dict(zip('ATCGatcg', 'TAGCtagc'))

def dna_revcomp(seq):
    """
    >>> dna_revcomp('ATG-CgcatN')
    'NatgcG-CAT'
    """
    return ''.join(_dna_rev_maps.get(c, c) for c in reversed(seq))

_colored = {
        'A': color_term('A', 'green'),
        'C': color_term('C', 'cyan'),
        'G': color_term('G', 'yellow'),
        'T': color_term('T', 'red'),
        'a': color_term('a', 'green'),
        'c': color_term('c', 'cyan'),
        'g': color_term('g', 'yellow'),
        't': color_term('t', 'red'),
}
def color_term_dna(seq):
    return ''.join(_colored.get(a, a) for a in seq)

# build codon table
STOP_AA = '$'
_aa_rnas = {
        'F': ('UUU', 'UUC',),
        'L': ('UUA', 'UUG', 'CUU', 'CUC', 'CUA', 'CUG',),
        'I': ('AUU', 'AUC', 'AUA',),
        'M': ('AUG',),                                     # Methionine start aa
        'V': ('GUU', 'GUC', 'GUA', 'GUG',),
        'S': ('UCU', 'UCC', 'UCA', 'UCG', 'AGU', 'AGC'),
        'P': ('CCU', 'CCC', 'CCA', 'CCG',),
        'T': ('ACU', 'ACC', 'ACA', 'ACG',),
        'A': ('GCU', 'GCC', 'GCA', 'GCG',),
        'Y': ('UAU', 'UAC',),
        STOP_AA: ('UAA', 'UAG', 'UGA',),
        'H': ('CAU', 'CAC',),
        'Q': ('CAA', 'CAG',),
        'N': ('AAU', 'AAC',),
        'K': ('AAA', 'AAG',),
        'D': ('GAU', 'GAC',),
        'E': ('GAA', 'GAG',),
        'C': ('UGU', 'UGC',),
        'W': ('UGG',),
        'R': ('CGU', 'CGC', 'CGA', 'CGG', 'AGA', 'AGG',),
        'G': ('GGU', 'GGC', 'GGA', 'GGG',),
}
_rna_aas = {}
_dna_aas = {}
for _aa, _rnas in _aa_rnas.items():
    for _rna in _rnas:
        _rna_aas[_rna] = _aa
        _dna_aas[_rna.replace('U', 'T')] = _aa

# https://en.wikipedia.org/wiki/Amino_acid
_AminoAcid = namedtuple('AminoAcid', 'name,name3,name1,side_chains,polarity,charge,h_index,mass')
amino_acids = [
        _AminoAcid('Alanine',        'Ala', 'A', ('aliphatic',),           'nonpolar',    0., 1.8, 89.094),
        _AminoAcid('Arginine',       'Arg', 'R', ('basic',),               'basic polar', 1., -4.5, 174.203),
        _AminoAcid('Asparagine',     'Asn', 'N', ('acid', 'amide'),        'polar',       0., -3.5, 132.119),
        _AminoAcid('Asparatic acid', 'Asp', 'D', ('acid', 'amide'),        'acidic polar', -1., -3.5, 133.104),
        _AminoAcid('Cysteine',       'Cys', 'C', ('sulfur-containing',),   'polar',       0., 2.5, 121.154),
        _AminoAcid('Glutamic acid',  'Glu', 'E', ('acid', 'amide'),        'acidic polar', -1., -3.5, 147.131),
        _AminoAcid('Glutamine',      'Gln', 'Q', ('acid', 'amide'),        'polar',       0., -3.5, 146.146),
        _AminoAcid('Glycine',        'Gly', 'G', ('aliphatic',),           'nonpolar',    0., -0.4, 75.067),
        _AminoAcid('Histidine',      'His', 'H', ('basic',),               'basic polar', .9, -3.2, 155.156),
        _AminoAcid('Isoleucine',     'Ile', 'I', ('aliphatic',),           'nonpolar',    0., 4.5, 131.175),
        _AminoAcid('Leucine',        'Leu', 'L', ('aliphatic',),           'nonpolar',    0., 3.8, 131.175),
        _AminoAcid('Lysine',         'Lys', 'K', ('basic',),               'basic polar', 1., -3.9, 146.189),
        _AminoAcid('Methionine',     'Met', 'M', ('sulfur-containing',),   'nonpolar',    0., 1.9, 149.208),
        _AminoAcid('Phenylalanine',  'Phe', 'F', ('aromatic',),            'nonpolar',    0., 2.8, 165.192),
        _AminoAcid('Proline',        'Pro', 'P', ('cyclic',),              'nonpolar',    0., -1.6, 115.132),
        _AminoAcid('Serine',         'Ser', 'S', ('hydroxyl-containing',), 'polar',       0., -0.8, 105.093),
        _AminoAcid('Theonine',       'Thr', 'T', ('hydroxyl-containing',), 'polar',       0., -0.7, 119.12),
        _AminoAcid('Tryptophan',     'Trp', 'W', ('aromatic',),            'nonpolar',    0., -0.9, 204.228),
        _AminoAcid('Tyrosine',       'Tyr', 'Y', ('aromatic',),            'polar',       0., -1.3, 118.191),
        _AminoAcid('Valine',         'Val', 'V', ('aliphatic',),           'nonpolar',    0., 4.2, 117.148),
]

def dna_translate(seq):
    """
    >>> dna_translate('ACTGGTTGCTAT')
    'TGCY'
    """
    return ''.join(_dna_aas[t1 + t2 + t3] for t1, t2, t3 in chunked(seq.upper(), 3))


def mask_same_bases(seq, ref, mask_char='.'):
    """
    >>> mask_same_bases('ATTG', 'AT-C')
    '..TG'
    """
    return ''.join((mask_char if s == r else s) for s, r in zip(seq, ref))

_pat_lclip = re.compile('^(-*)[^-]')
_pat_rclip = re.compile('.*[^-](-*)$')

def get_lclip(seq):
    """
    >>> get_lclip('---AG---')
    3
    >>> get_lclip('A---')
    0
    >>> get_lclip('---')
    0
    """
    m = _pat_lclip.match(seq)
    return len(m.group(1)) if m else 0

def get_rclip(seq):
    """
    >>> get_rclip('---AG---')
    3
    >>> get_rclip('A---')
    3
    >>> get_rclip('---')
    0
    """
    m = _pat_rclip.match(seq)
    return len(m.group(1)) if m else 0

_unknown_pat = re.compile('-*[Nn]+-*')
def mask_unknown_seq(seq, mask_char='*'):
    """
    >>> aln = '----------AGGGNNG-TT'
    >>> aln
    '----------AGGGNNG-TT'

    >>> mask_unknown_seq(aln)
    '**********AGGG**G-TT'

    >>> aln = '----------AGGG-GTT---NN--NNNNNNN-NNNN--C-CCTA---------TTANNNN--NNNNNNNN-----'
    >>> aln
    '----------AGGG-GTT---NN--NNNNNNN-NNNN--C-CCTA---------TTANNNN--NNNNNNNN-----'

    >>> mask_unknown_seq(aln)
    '**********AGGG-GTT*********************C-CCTA---------TTA*******************'
    """
    prev_end = 0
    tokens = []
    seq1 = ''.join(('N', seq, 'N'))
    for m in _unknown_pat.finditer(seq1):
        start = m.start()
        end = m.end()
        tokens.append(seq1[prev_end:start])
        tokens.append(mask_char * (end - start))
        prev_end = end

    tokens.append(seq1[prev_end:])
    return ''.join(tokens)[1:-1]  # remove ends


_SeqPoint = namedtuple('SeqPoint', 'is_block index inner_pos')
_SeqInterval = namedtuple('SeqInterval', 'is_block index start end')

class SeqLocator(object):
    """
    pos is 0-based coordinate

    last pos
    0 1 1
    : : :
    |A|-|
                         1         2         3         4         5         6         7
    pos        0123456789012345678901234567890123456789012345678901234567890123456789012345
    is_block   0         1       0                    1                 0
    index      0         0       1                    1                 2
    block_id             0                            1
    last_end             01234456                     011234555555555567
                                                           1
    cum_last_end         01234456                     677890111111111123
    >>> aln = '**********AGGG-GTT*********************C-CCTA---------TTA*******************'
    >>> loc = SeqLocater(aln)
    >>> loc.blocks
    [(10, 18, 'AGGG-GTT'), (39, 57, 'C-CCTA---------TTA')]

    >>> loc.get_context(0)
    (None, None, '*')
    >>> loc.get_context(9, left=3, right=3)
    (None, None, '******')
    >>> loc.get_context(10, left=3, right=3)
    (0, 0, '***AGG')
    >>> loc.get_context(13, left=3, right=3)
    (0, 3, 'AGGGGT')
    >>> loc.get_context(14, left=3, right=3)
    (0, 4, 'GGGGTT')
    >>> loc.get_context(45, left=4, right=4)
    (1, 5, 'CCTATTA*')
    >>> loc.get_context(56, left=3, right=3)
    (1, 7, 'ATTA**')

    >>> [tuple(iv) for iv in loc.intervals]   # blank or block intervals
    [(0, 0, 0, 10), (1, 0, 10, 18), (0, 1, 18, 39), (1, 1, 39, 57), (0, 2, 57, 76)]

    >>> x = loc.get_point(9)
    >>> (x.is_block, x.index, x.inner_pos)
    (0, 0, 9)

    >>> x = loc.get_point(13)
    >>> (x.is_block, x.index, x.inner_pos)
    (1, 0, 3)
    """
    _mask_char = '*'
    _mask_pat = re.compile('\*+')
    def __init__(self, aln):
        self.aln = aln
        prev_end = 0
        blocks = []   # [(s, e, aln)]
        for m in self._mask_pat.finditer(aln):
            start = m.start()
            end = m.end()
            if start > prev_end:
                blocks.append((prev_end, start, aln[prev_end:start]))
            prev_end = end
        if prev_end < len(self.aln):
            blocks.append((prev_end, len(self.aln), aln[prev_end:]))
        self.blocks = blocks

        # add intervals
        self.intervals = ivs = []   # [SeqInterval]
        prev_end = 0
        for idx, b in enumerate(blocks):
            ivs.append(_SeqInterval(0, idx, prev_end, b[0]))  # not a block
            ivs.append(_SeqInterval(1, idx, b[0], b[1]))      # a block
            prev_end = b[1]
        ivs.append(_SeqInterval(0, len(blocks), prev_end, len(aln)))   # last blank

        self._pos_maps = []
        for s, e, aln in blocks:
            is_regular = np.array(list(aln)) != '-'
            pos_map = np.cumsum(is_regular) - is_regular
            self._pos_maps.append(pos_map)
            logging.debug(pos_map)

    def get_point(self, pos):   # linear search
        """
        Returns: SeqPoint
        """
        assert 0 <= pos < len(self.aln)
        for iv in self.intervals:
            if pos < iv.end:
                offset = pos - iv.start
                return _SeqPoint(iv.is_block, iv.index, offset)
        raise Exception('This should not be called')

    def get_block_index(self, pos):  # linear search
        assert 0 <= pos < len(self.aln)
        for idx, block in enumerate(self.blocks):
            if pos < block[0]:
                break
            if pos < block[1]:
                return idx
        return None

    def get_context(self, pos, left=0, right=1):
        """ Returns: (block_index, block_last-end, context)
        """
        assert left >= 0 and right >= 0
        idx = self.get_block_index(pos)
        if idx is None:
            return (None, None, self._mask_char * left + self._mask_char * right)
        start, end, aln = self.blocks[idx]
        pos_map = self._pos_maps[idx]
        offset = pos - start
        block_last_end = pos_map[offset]
        lseq = aln[:offset].replace('-', '')
        rseq = aln[offset:].replace('-', '')
        lseq1 = (self._mask_char * left + lseq)[-left:] if left > 0 else ''
        rseq1 = (rseq + self._mask_char * right)[:right]
        return (idx, block_last_end, lseq1 + rseq1)

# deprecated
SeqLocater = SeqLocator

def get_aln_pos_text(aln, offset=0, digit=5, count_del=False):
    """
    Args:
        start 0-based coordinate

    >>> offset = 10010054
    >>> t = get_aln_pos_text(('AAAAA' 'TTTTT' 'GGGGG' 'CCCCC'), offset=offset)
    >>> t['number']
    '      10061     10071'
    >>> t['indicator']
    '      |         |   '

    >>> t = get_aln_pos_text(('AA--AAA' 'TTTTT' 'G--GGGG' 'CCCCC'), offset=offset)
    >>> t['number']
    '        10061       10071'
    >>> t['indicator']
    '        |           |   '

    >>> t = get_aln_pos_text(('AA--AAA' 'TTTTT' 'G--GGGG' 'CCCCC'), offset=offset, count_del=True)
    >>> t['number']
    '      10061     10071   '
    >>> t['indicator']
    '      |         |       '

    >>> t = get_aln_pos_text('AAAAA', offset=offset)
    >>> t['number']
    '10055'
    >>> t['indicator']
    '|    '
    """
    poss = []  # poss to print
    offsets = []
    step = 10
    j = left = offset % step
    for i in xrange(len(aln)):
        b = aln[i]
        if count_del or b != '-':
            j += 1
            if j % step == 1:
                poss.append(offset - left + j)
                offsets.append(i)
    if not poss:
        poss.append(offset + 1)
        offsets.append(0)
    offsets.append(len(aln))   # last offset

    n = 10 ** digit
    o = offsets[0]
    buf_n = [' ' * o]
    buf_i = [' ' * o]
    for p, o1, o2 in zip(poss, offsets, offsets[1:]):
        num = str(p % n)
        o = o2 - o1
        buf_n.append(fill_text(num, o))
        buf_i.append(fill_text('|', o))
    return {'number': ''.join(buf_n), 'indicator': ''.join(buf_i)}


ChainHeader = namedtuple('ChainHeader', 'score tName tSize tStrand tStart tEnd qName qSize qStrand qStart qEnd id')

def contig2chain(from_contig, to_contig, id=1):
    """ Convert bed to chain file

    Args:
        bed: iterable of (chrom, start, end, length)
        rlenths : dictionary of {rname: length}
    Return:
        {'header': header tokens
         'data': [(size, dt, dq), ... , (size,)}

    See http://genome.ucsc.edu/goldenPath/help/chain.html for more details.

    header tokens:
        score -- chain score
        tName -- chromosome (reference sequence)
        tSize -- chromosome size (reference sequence)
        tStrand -- strand (reference sequence)
        tStart -- alignment start position (reference sequence)
        tEnd -- alignment end position (reference sequence)
        qName -- chromosome (query sequence)
        qSize -- chromosome size (query sequence)
        qStrand -- strand (query sequence)
        qStart -- alignment start position (query sequence)
        qEnd -- alignment end position (query sequence)
        id -- chain ID

    data tokens (the end of data token contains only first token):
        size -- the size of the ungapped alignment
        dt -- the difference between the end of this block and the beginning of the next block (reference sequence)
        dq -- the difference between the end of this block and the beginning of the next block (query sequence)

    >>> reader = ('>contig1', 'AAA-TTT--ATTTTGT', '>contig2', 'A---CGTATATTTGG-')
    >>> fasta = Fasta(reader)
    >>> result = contig2chain(fasta.get('contig1'), fasta.get('contig2'))
    >>> attrgetter('score', 'tName', 'tSize', 'tStart', 'tEnd', 'qName', 'qSize', 'qStart', 'qEnd', 'id')(result['header'])
    (0, 'contig1', 13, 0, 13, 'contig2', 12, 0, 12, 1)
    >>> result['data']
    [[1, 2, 0], [3, 0, 2], [6, 1, 0], [0]]
    """
    from_size = len(from_contig.seq.replace('-', ''))
    to_size = len(to_contig.seq.replace('-', ''))
    header = ChainHeader(
            score=0,
            tName=from_contig.name, tSize=from_size, tStrand='+', tStart=0, tEnd=from_size,
            qName=to_contig.name, qSize=to_size, qStrand='+', qStart=0, qEnd=to_size,
            id=id,
    )
    data = []
    states = []
    def iter_states(seq1, seq2):
        for s1, s2 in zip(seq1, seq2):
            if s1 == s2 == '-':
                yield None
            elif s1 == '-':
                yield 2
            elif s2 == '-':
                yield 1
            else:
                yield 0

    it = iter_states(from_contig.seq, to_contig.seq)
    it = (el for el in it if el is not None)
    it = ((st, count) for st, count in iter_counts(it))

    dat = []
    for st, count in it:
        if st == 0:
            dat.append([count])
        elif st == 1:
            if not dat:
                dat.append([0])
            assert len(dat[-1]), 'Unexpected last data: {0} for state: {1}'.format(dat, st)
            dat[-1].extend((count, 0))
        elif st == 2:
            if not dat:
                dat.append([0])
            assert len(dat[-1]), 'Unexpected last data: {0} for state: {1}'.format(dat, st)
            dat[-1].extend((0, count))
        else:
            raise Exception('Unexpected state: {0}'.format(st))

    if not dat or len(dat[-1]) == 3:
        dat.append([0])

    return {'header': header, 'data': dat}


#TODO use .fai file if possible
class Fasta(object):
    """
    >>> reader = ('>contig1', 'AAATTT', '>contig2', 'ACGT')
    >>> fasta = Fasta(reader)
    >>> fasta.get('contig2').seq
    'ACGT'
    """

    def __init__(self, reader):
        self._reader = iter(reader)
        self._iterator_called = False

    @cached_property
    def contigs(self):
        assert self._iterator_called == False, 'Do not iterate before call the method'
        return list(self)

    @cached_property
    def _name_contigs(self):
        return dict((c.name, c) for c in self.contigs)

    def __contains__(self, name):
        return name in self._name_contigs

    def get(self, name):
        return self._name_contigs[name]

    @property
    def names(self):
        return [contig.name for contig in self.contigs]

    @property
    def lengths(self):
        return [len(contig) for contig in self.contigs]

    def __iter__(self):
        self._iterator_called = True
        contig = None
        try:
            while self._reader:
                line = next(self._reader).rstrip()
                if line.startswith('>'):
                    if contig is not None:
                        yield contig
                    contig = FastaContig(line[1:])
                else:
                    contig.append(line)
        except StopIteration:
            if contig is not None:
                yield contig


class FastaContig(object):
    """
    >>> contig = FastaContig('contig description')
    >>> contig.name_line
    'contig description'
    >>> contig.name
    'contig'
    >>> contig.append('AAAA')
    >>> contig.append('TTTT')
    >>> contig.seq
    'AAAATTTT'
    >>> contig.append('GGGG')
    >>> contig.seq
    'AAAATTTTGGGG'
    >>> len(contig)
    12
    >>> contig.lclip
    0

    >>> contig.append('----')
    >>> contig.seq
    'AAAATTTTGGGG----'
    >>> contig.rclip
    4

    >>> contig.get_clip_mask_seq()
    'AAAATTTTGGGG****'
    >>> contig.append('C-CC----')
    >>> contig.get_clip_mask_seq()
    'AAAATTTTGGGG----C-CC****'

    >>> contig.prepend('----')
    >>> contig.get_clip_mask_seq()
    '****AAAATTTTGGGG----C-CC****'
    """

    def __init__(self, name_line=''):
        assert not name_line.startswith('>'), 'please remove charcter ">"'
        self.name_line = name_line
        self.name = name_line.split(' ', 2)[0]
        self._seqs = []

    def prepend(self, seq):
        self._seqs.insert(0, seq)
        self.__dict__.pop('seq', None)

    def append(self, seq):
        self._seqs.append(seq)
        self.__dict__.pop('seq', None)

    @cached_property
    def seq(self):
        return ''.join(self._seqs)

    def __len__(self):
        return len(self.seq)

    def get_clip_mask_seq(self, mask_char='*'):
        seq = mask_char * self.lclip + self.seq[self.lclip : len(self) - self.rclip] + mask_char * self.rclip
        return seq

    @property
    def lclip(self):
        return get_lclip(self.seq)

    @property
    def rclip(self):
        return get_rclip(self.seq)


class AlignmentAnalyzer(object):
    """
    0-based pos

    pos         01234567
    last_end1   01234456
    last_end2   01233345
    >>> aln1 = 'AAAA-AAC'
    >>> aln2 = 'AAA--ACT'
    >>> aa = AlignmentAnalyzer(aln1, aln2)
    >>> list(aa.iter_edits())
    [(3, 'A', '-'), (6, 'A', 'C'), (7, 'C', 'T')]

    >>> keys = 'last_end1 ctx1 last_end2 ctx2'.split(' ')
    >>> ctx = aa.get_context(3, left=4, right=4)
    >>> [ctx[k] for k in keys]
    [3, '*AAAAAAC', 3, '*AAAACT*']

    >>> ctx = aa.get_context(6, left=4, right=4)
    >>> [ctx[k] for k in keys]
    [5, 'AAAAAC**', 4, 'AAAACT**']
    """

    def __init__(self, aln1, aln2, validate=True):
        if validate:
            assert len(aln1) == len(aln2)
            assert aln1 == aln1.upper()
            assert aln2 == aln2.upper()
        self.aln1 = aln1
        self.aln2 = aln2
        _aln1 = np.array(list(aln1))
        _aln2 = np.array(list(aln2))
        self._is_edit = (_aln1 != _aln2)

    @cached_property
    def _loc1(self):
        return SeqLocator(self.aln1)

    @cached_property
    def _loc2(self):
        return SeqLocator(self.aln2)

    def iter_edits(self):
        """ (pos_msa, aln1, aln2)
        """
        for pos in np.argwhere(self._is_edit).flatten():
            yield (pos, self.aln1[pos], self.aln2[pos])

    def get_context(self, pos, left=0, right=1):
        ctxs = {}
        ctxs['block_idx1'], ctxs['last_end1'], ctxs['ctx1'] = self._loc1.get_context(pos, left=left, right=right)
        ctxs['block_idx2'], ctxs['last_end2'], ctxs['ctx2'] = self._loc2.get_context(pos, left=left, right=right)
        return ctxs



from __future__ import absolute_import
from __future__ import print_function
import sys
import logging
from argtools import command, argument
from itertools import groupby
from collections import defaultdict
from .. import sh
import vcf
import cyvcf2 as VCF
import numpy as np
from ..utils import memoize, iter_tabs


@command.add_sub
@argument('vcf', help='VCF or gzipped VCF')
def vcf_header(args):
    vcffile = '/dev/stdin' if args.vcf == '-' else args.vcf
    with open(vcffile) as fp:
        r = vcf.Reader(fp)
        print ('category', 'id', 'num', 'type', 'desc', sep='\t')
        for (key, rec) in r.filters.items():
            print ('FILTER', rec.id, '', '', rec.desc, sep='\t')

        for (key, rec) in r.infos.items():
            print ('INFO', rec.id, rec.num, rec.type, rec.desc, sep='\t')

        for (key, rec) in r.formats.items():
            print ('FORMAT', rec.id, rec.num, rec.type, rec.desc, sep='\t')

        for (key, rec) in r.alts.items():
            print ('ALT', rec.id, '', '', rec.desc, sep='\t')

        for (key, rec) in r.metadata.items():
            print ('METADATA', key, '', '', rec, sep='\t')


@command.add_sub
@argument('vcf', help='VCF or gzipped VCF')
def vcf_samples(args):
    vcffile = '/dev/stdin' if args.vcf == '-' else args.vcf
    with open(vcffile) as fp:
        r = vcf.Reader(fp)
        for sample in r.samples:
            print (sample, sep='\t')


def _get_sample_flags(reader, args):
    sample_set = set()
    if args.samples:
        sample_set = set(args.samples.split(','))
    if args.samples_file:
        sample_set = set([line.rstrip() for line in open(args.samples_file)])

    if sample_set:
        sample_flags = np.array([s in sample_set for s in reader.header.samples], dtype=bool)
    else:
        sample_flags = np.ones(len(reader.header.samples), dtype=bool)

    return sample_flags


@memoize
def get_gt_hom_idxs(alt_num):
    """
    >>> get_gt_hom_idxs(0)  # 0/0
    [0]
    >>> get_gt_hom_idxs(1)  # 0/0, 0/1, 1/1
    [0, 2]
    >>> get_gt_hom_idxs(2)  # 0/0, 0/1, 1/1, 0/2, 1/2, 2/2
    [0, 2, 5]
    >>> get_gt_hom_idxs(3)
    [0, 2, 5, 9]
    """
    last = -1
    hom_idxs = []
    for a in range(alt_num + 1):
        last = last + (a + 1)
        hom_idxs.append(last)
    return hom_idxs


@memoize
def get_default_gl(alt_num):
    """
    >>> get_default_gl(1)
    '-0.301,0,-0.301'
    >>> get_default_gl(2)
    '-0.301,0,-0.301,0,0,-0.301'
    """
    het_gl = '0'
    hom_gl = '{0:.3f}'.format(- np.log10(2))
    gt_num = (alt_num + 1) * (alt_num + 2) / 2
    gl = [het_gl] * gt_num
    hom_idxs = get_gt_hom_idxs(alt_num)
    for idx in hom_idxs:
        gl[idx] = hom_gl
    return ','.join(gl)


@command.add_sub
@argument('vcf', help='VCF or gzipped VCF')
@argument.exclusive(
    argument('-s', '--samples', help='comma separated sample list'),
    argument('-S', '--samples-file', help='sample list file'),
)
def vcf_reset_gl(args):
    """ whitening format tag for specified samples

    # TODO also reset PL
    """
    vcffile = '/dev/stdin' if args.vcf == '-' else args.vcf
    import pysam
    reader = pysam.VariantFile(vcffile)
    sample_flags = _get_sample_flags(reader, args)
    print (reader.header, end='')
    sample_offset = 9

    # hom: log10 p**2 = 2 log10 p
    # het: log10 2p**2 = 2 log10 p + log10 2

    def replace_fmt(alt_num, gl_index, fmt_info):
        splt = fmt_info.split(':')
        splt[gl_index] = get_default_gl(gt_num)
        return ':'.join(splt)

    for rec in reader:
        srec = str(rec).split('\t')
        fmts = rec.format.keys()
        alt_num = len(rec.alts)
        gl_index = fmts.index('GL')

        fmt_rec = [(replace_fmt(alt_num, gl_index, s) if flg else s) for flg, s in zip(sample_flags, srec[sample_offset:])]
        print (*(srec[:sample_offset] + fmt_rec), sep='\t', end='')


@command.add_sub
@argument('vcf', help='VCF or gzipped VCF')
@argument.exclusive(
    argument('-s', '--samples', help='comma separated sample list'),
    argument('-S', '--samples-file', help='sample list file'),
)
def vcf_untype(args):
    """ whitening format tag for specified samples

    # TODO also reset PL
    """
    vcffile = '/dev/stdin' if args.vcf == '-' else args.vcf
    import pysam
    reader = pysam.VariantFile(vcffile)
    sample_flags = _get_sample_flags(reader, args)
    print (reader.header, end='')
    sample_offset = 9

    import re
    gt_sep = re.compile('/|')
    def replace_fmt(gt_index, fmt_info):
        splt = fmt_info.split(':')
        gt = splt[gt_index]
        ngt = len(gt_sep.split(gt))
        splt[gt_index] = '/'.join(('.',) * ngt)
        return ':'.join(splt)

    for rec in reader:
        srec = str(rec).split('\t')
        fmts = rec.format.keys()
        gt_index = fmts.index('GT')

        fmt_rec = [(replace_fmt(gt_index, s) if flg else s) for flg, s in zip(sample_flags, srec[sample_offset:])]
        print (*(srec[:sample_offset] + fmt_rec), sep='\t', end='')


def _load_pos_sets(poslist):
    chrom_pos_sets = defaultdict(set)
    with open(poslist) as fp:
        for row in iter_tabs(fp):
            chrom = row[0]
            pos = int(row[1])
            chrom_pos_sets[chrom].add(pos)
    return chrom_pos_sets


@command.add_sub
@argument('vcf', help='VCF or gzipped VCF')
@argument('poslist', help='chrom:pos')
@argument('-r', '--region', help='VCF or gzipped VCF')
def vcf_pick(args):
    """ poslist: chrom:pos
    """
    vcffile = '/dev/stdin' if args.vcf == '-' else args.vcf
    import pysam
    reader = pysam.VariantFile(vcffile)

    # iteration
    if args.region:
        it = reader.fetch(region=args.region)
    else:
        it = reader

    logging.info('Loading poslist: %s', args.poslist)
    chrom_pos_sets = _load_pos_sets(args.poslist)
    for chrom in sorted(chrom_pos_sets):
        logging.info('Loaded %s pos from chrom: %s', len(chrom_pos_sets[chrom]), chrom)

    print (reader.header, end='')
    for chrom, recs in groupby(it, lambda x: x.contig):
        pos_sets = chrom_pos_sets[chrom]
        for rec in recs:
            if rec.pos in pos_sets:
                print (str(rec), end='')


@command.add_sub
@argument('vcf', help='VCF or gzipped VCF')
@argument('-r', '--region', help='VCF or gzipped VCF')
@argument('-c', '--check-variant', action='store_true', help='consider duplicated if check ref and alts are concordant')
def vcf_list_dup(args):
    """
    Emit:
        chrom, pos, id, ref, alt
    """
    vcffile = '/dev/stdin' if args.vcf == '-' else args.vcf
    import pysam
    reader = pysam.VariantFile(vcffile)

    # iteration
    if args.region:
        it = reader.fetch(region=args.region)
    else:
        it = reader

    if args.check_variant:
        def get_id(rec):
            return ':'.join((rec.chrom, str(rec.pos), rec.ref, ','.join(rec.alts)))
    else:
        def get_id(rec):
            return ':'.join((rec.chrom, str(rec.pos)))

    print ('chrom', 'pos', 'id', 'ref', 'alt', sep='\t')
    checked = set()  # pos ':' ref ':' ','.join(alts)
    for chrom, recs in groupby(it, lambda x: x.contig):
        last_pos = 0
        for rec in recs:
            rec_id = get_id(rec)
            if rec_id in checked:
                print (rec.chrom, rec.pos, rec.id, rec.ref, ','.join(rec.alts), sep='\t')
                continue
            checked.add(rec_id)



from builtins import filter, zip
from collections import namedtuple, defaultdict
from itertools import takewhile, dropwhile, chain, groupby
from .bioseq import dna_revcomp
from .samutil import alignment_to_cigar, CIGAR
import pysam
import numpy as np
import pandas as pd


class BlastTabFile:
    # Fields: query id, subject id, % identity, alignment length, mismatches, gap opens, q. start, q. end, s. start, s. end, evalue, bit score, gaps, query length, subject length, query seq, subject seq
    _field_keys = {
            'query id': 'qname',
            'subject id': 'sname',
            '% identity': 'identity',
            'alignment length': 'aln_len',
            'mismatches': 'mismatches',
            'gap opens': 'gap_opens',
            'q. start': 'qstart',
            'q. end': 'qend',
            's. start': 'sstart',
            's. end': 'send',
            'evalue': 'evalue',
            'bit score': 'bit_score',
            'gaps': 'gaps',
            'query length': 'qlen',
            'subject length': 'slen',
            'query seq': 'qseq',
            'subject seq': 'sseq',
            }

    def __init__(self, reader):
        # read header
        it = iter(dropwhile(lambda x: not x.startswith('# Fields'), reader))
        line = next(it)    # Fields
        self.fields = [f.strip() for f in line.split(':', 1)[1].split(',')]
        self.keys = [self._field_keys[f] for f in self.fields]  # blast key
        #self._convs = [self._types.get(k) for k in self.keys]
        BlastRecord = namedtuple('BlastRecord', self.keys)
        #class BlastRecord(namedtuple('BlastRecord', self.keys), BlastRecordMixin):
        #    pass

        self._Record = BlastRecord
        self._reader = it

    def iter_raw(self):
        for line in self._reader:
            if line.startswith('#'):
                continue
            rec = line.rstrip().split('\t')
            #rec = [rec if conv is None else conv(rec) for conv, rec in zip(self._convs, rec)]
            yield self._Record(*rec)

    def iter_query_tables(self, analyze=True):
        from cStringIO import StringIO
        it = self._reader
        while 1:
            output = StringIO()
            it = takewhile(lambda x: not x.startswith('#'),
                    dropwhile(lambda x: x.startswith('#'), self._reader))
            for line in it:
                output.write(line)

            output.seek(0)
            tab = pd.read_table(output, names=self.keys, index_col=False)
            #logging.debug(tab)
            if not len(tab):
                break
            qname = tab['qname'][0]
            if analyze:
                self._analyze(tab)
            yield qname, tab
            output.close()

    @staticmethod
    def _analyze(tab):
        #logging.debug(tab)
        tab['edit_distance'] = tab['mismatches'] + tab['gaps']
        tab['qratio'] = 1. * ((tab['qend'] - tab['qstart']) + 1) / tab['qlen']
        tab['sratio'] = 1. * ((tab['send'] - tab['sstart']).abs() + 1) / tab['slen']
        tab['qlclip'] = tab['qstart'] - 1
        tab['qrclip'] = tab['qlen'] - tab['qend']
        tab['is_reverse'] = tab['send'] < tab['sstart']
        tab['slclip'] = np.where(tab['is_reverse'], tab['slen'] - tab['send'], tab['sstart'] - 1)
        tab['srclip'] = np.where(~tab['is_reverse'], tab['sstart'] - 1, tab['slen'] - tab['send'])
        tab['left_overhang'] = np.min([tab.qlclip, tab.slclip], axis=0)
        tab['right_overhang'] = np.min([tab.qrclip, tab.srclip], axis=0)

    def iter_sam_record(self, reference_ids, tags=None):
        mapq = 30
        bq = '+'
        it = self.iter_raw()

        # TODO secondary alignment flag
        for rec in it:
            is_reverse = int(rec.send) < int(rec.sstart)
            if is_reverse:
                qaln = dna_revcomp(rec.qseq)
                saln = dna_revcomp(rec.sseq)
                ref_start = int(rec.send) - 1
                lclip = int(rec.qstart) - 1
                rclip = int(rec.qend) - int(rec.qend)
            else:
                qaln = rec.qseq
                saln = rec.sseq
                ref_start = int(rec.sstart) - 1
                lclip = int(rec.qlen) - int(rec.qend)
                rclip = int(rec.qstart) - 1
            cigar = alignment_to_cigar(qaln, saln)
            qseq = qaln.replace('-', '')
            if lclip > 0:
                cigar.insert(0, (CIGAR.H, lclip))
            if rclip > 0:
                cigar.append((CIGAR.H, rclip))

            try:
                edit = int(rec.mismatches) + int(rec.gaps)
            except:
                edit = -1

            a = pysam.AlignedSegment()
            a.query_name = rec.qname.encode('ascii')
            a.query_sequence = qseq.encode('ascii')
            a.reference_id = reference_ids[rec.sname]
            a.flag = (16 if is_reverse else 0)
            a.reference_start = ref_start
            a.mapping_quality = mapq
            a.cigar = cigar
            a.next_reference_id = -1
            a.next_reference_start = -1
            try:
                a.template_length = int(rec.slen)
            except:
                pass
            a.query_qualities = pysam.qualitystring_to_array(bq * len(qseq))
            a.tags = [
                    ("AS", float(rec.bit_score)), # alignment score
                    ("NM", edit),               # edit distance
                    ("ZE", float(rec.evalue)),  # E-value
            ]
            yield a



from __future__ import print_function
from argtools import command, argument
from itertools import dropwhile, chain, groupby
from builtins import filter, zip
from collections import namedtuple, defaultdict
import logging
import re
import os
import time
from .. import sh
from ..utils import cached_property, collect_while, skip_until, isblank, make_not, blank_split
from ..bioseq import dna_revcomp, dna_translate
from ..blast import BlastTabFile
import pandas as pd
import numpy as np
from joblib import Parallel, delayed


class IMGTNucFile:
    '''
    Example:
        <header>

        HLA-A Nucleotide Sequence Alignments
        IMGT/HLA Release: 3.24.0.1
        Sequences Aligned: 2016 May 04
        Steven GE Marsh, Anthony Nolan Research Institute.
        Please see http://hla.alleles.org/terms.html for terms of use.


    '''
    @staticmethod
    def _parse_lines(nuc_file):
        with open(nuc_file) as fp:
            try:
                it = fp
                headers, it = collect_while(make_not(isblank), it)
                bodies = []
                logging.info('Read headers')

                # body
                while 1:
                    data = {}
                    _, it = collect_while(isblank, it)  # skip blank lines

                    line = next(it).rstrip()
                    logging.debug(line)
                    tokens = blank_split(line)
                    if tokens[0] != 'cDNA':
                        raise StopIteration
                    data['dna_poss'] = list(map(int, tokens[1:]))

                    line = next(it).rstrip()
                    logging.debug(line)
                    if line.strip().startswith('AA'):
                        tokens = blank_split(line)
                        assert tokens[0] == 'AA' and tokens[1] == 'codon'
                        data['aa_poss'] = list(map(int, tokens[2:]))
                        line = next(it).rstrip()
                        logging.debug(line)

                    data['offsets'] = [i for i, s in enumerate(line) if s == '|']

                    data['lines'], it = collect_while(make_not(isblank), it)
                    logging.debug('%s', data['lines'][0].rstrip())
                    if len(data['lines']) > 1:
                        logging.debug('%s', data['lines'][1].rstrip())
                    logging.debug('Read %s subtypes', len(data['lines']))
                    bodies.append(data)
            except StopIteration as e:
                pass
            logging.info('Read bodies')
        return headers, bodies

    def __init__(self, nuc_file):
        headers, bodies = self._parse_lines(nuc_file)
        # headers
        self.name = headers[0].rstrip()
        self.gene = self.name.split(' ')[0]  # HLA-A
        self.version = headers[1].rstrip().split(':')[1].strip()  # 3.24.0.1
        # bodies
        self.gene_info = []
        offsets = []
        self.subtypes = subs = [blank_split(line)[0] for line in bodies[0]['lines']]
        sub_alns = dict((subtype, []) for subtype in subs)  # {subtype: [exon alignment]}
        for body in bodies:
            offsets.extend(body['offsets'])

            # the first line of each body block is the refenence alignment
            line = body['lines'][0]
            tokens = blank_split(line)
            subtype = tokens[0]
            ref_aln = ''.join(tokens[1:])
            sub_alns[subtype].append(ref_aln)
            for line in body['lines'][1:]:
                tokens = blank_split(line)
                subtype = tokens[0]
                raw_aln = ''.join(tokens[1:])
                aln = decode_seq(raw_aln, ref_aln)
                sub_alns[subtype].append(aln)

        logging.info('Load %s subtypes', len(subs))
        logging.info('Alignment block length: %s', len(ref_aln))
        self.subtype_nucs = dict((sub, IMGTNuc(sub, ''.join(sub_alns[sub]))) for sub in subs)


def decode_seq(seq, ref, same='-'):
    """
    >>> decode_seq('-G-.-A.-*-', 'ATACTTTG..')
    'AGA.TA.G*.'
    """
    return ''.join(r if a == same else a for a, r in zip(seq, ref))

def get_exon_frames(exon_seqs):
    """
    >>> get_exon_frames(['***GC*', 'GA****'])
    [(0, 0), (0, 0)]
    >>> get_exon_frames(['AT', 'GCTTTA', 'AT'])
    [(0, 2), (2, 2), (2, 1)]
    >>> get_exon_frames(['ATT'])
    [(0, 0)]
    """
    frames = []
    right = 0
    for seq in exon_seqs:
        left = right
        right = (left + len(seq)) % 3
        frames.append((left, right))
    return frames

class IMGTNuc:
    def __init__(self, name, aln):
        """
        name: A*01:01:01:02N
        aln: ATGGCCGTCATGGCGCCCCGAACCCTCCTC...
        """
        raw_sep_seq = aln.replace('.', '')  # remove deletion char
        raw_exon_seqs = raw_sep_seq.split('|')  # exons with '*'
        exon_frames = get_exon_frames(raw_exon_seqs)

        self.seq = raw_sep_seq.replace('|', '')
        self.exons = raw_exon_seqs
        self.exon_frames = exon_frames


@command.add_sub
@argument('nuc')
def imgt_nuc2exon(args):
    parsed = IMGTNucFile(args.nuc)
    print ('subtype', 'number', 'is_typed', 'lphase', 'rphase', 'length', 'seq', sep='\t')
    for subtype in parsed.subtypes:
        nuc = parsed.subtype_nucs[subtype]
        for number, (seq, (lphase, rphase)) in enumerate(zip(nuc.exons, nuc.exon_frames), 1):
            is_typed = int('*' not in seq)
            print (subtype, number, is_typed, lphase, rphase, len(seq), seq, sep='\t')


#@command.add_sub
@argument('gen')
def imgt_gen(args):
    pass


#@command.add_sub
@argument('prot')
def imgt_prot(args):
    pass

@command.add_sub
@argument('contigs')
@argument('-d', '--imgt-dir', default=None)
@argument('-g', '--imgt-gen', default=None)
@argument('-l', '--allele-list', default=None)
@argument('-o', '--out-dir', default='imgt_classify_hla')
@argument('-t', '--target-genes', nargs='+', default=['A', 'B', 'C'])
@argument('-f', '--force-rerun', action='store_true')
@argument('--dry-run', action='store_true')
@argument('--EST2GENOME', default='est2genome')
@argument('-j', '--njobs', type=int, default=8)
def imgt_classify_hla(args):
    """
    Currently, classical imgt type only

    Inputs:
        {imgt_dir}/fasta/{gene}_nuc.fasta
        gen.fasta
        required files under {imgt_aln_dir}
        *_nuc.txt is required for identify the exon number contained in cDNA (*_nuc.fasta)
        *_nuc.txt

    Outputs:
        {outdir}/{gene}_nuc.exons.txt
        {outdir}/{gene}_nuc.fa
        {outdir}/{gene}_nuc.fa
        {outdir}/gen.blastdb
        {outdir}/contig.gen.blastn
        {outdir}/contig.gen.blastn.bests
        {outdir}/contigs/{contig_name}.fa
        {outdir}/contigs/{contig_name}.{gene}.est2genome
        {outdir}/contig_summary1.txt
        {outdir}/contig_summary2.txt

    # gene type
    # 4 digit (protein level)
    # 6 digit (exon level)     <= maybe detection of exon 8 will be failed
    # 6G digit (exon 2 and 3) wmda/hla_nom_G.txt
    # 8 digit (outside) if exists
    # variant on exon
    # variant on intron
    """
    contigs_fasta = args.contigs
    gen_fasta_org = args.imgt_gen
    imgt_dir = args.imgt_dir
    allele_list = args.allele_list
    out_dir = args.out_dir
    sh.dry_run = args.dry_run

    sh.call(['mkdir', '-p', out_dir])
    gen_blastdb = '{dir}/gen.blastdb'.format(dir=out_dir)
    gen_blastdb_nhr = gen_blastdb + '.nhr'

    gen_fasta = '{dir}/gen.fa'.format(dir=out_dir)
    contig_gen_blast = '{dir}/contigs.gen.blastn'.format(dir=out_dir)
    contig_gen_blast_bests = '{dir}/contigs.gen.blastn.bests'.format(dir=out_dir)
    contig_cdna_dir = '{dir}/contigs'.format(dir=out_dir)
    closest_exons = '{dir}/contig.closest_exons.txt'.format(dir=out_dir)
    closest_genomes = '{dir}/contig.closest_genomes.txt'.format(dir=out_dir)
    contig_summary = '{dir}/contig.summary.txt'.format(dir=out_dir)
    contig_nuc_fa_tmpl = '{dir}/contig.summary.{{gene}}_nuc.fa'.format(dir=out_dir)
    contig_gen_fa_tmpl = '{dir}/contig.summary.{{gene}}_gen.fa'.format(dir=out_dir)
    exon_counts = {'A': 8, 'B': 7, 'C': 8, 'H': 8}   # TODO more systematic way

    if imgt_dir is None:
        logging.error('imgt_dir is required for preprocss')
        raise Exception('No data directory')

    if args.force_rerun or not os.path.exists(gen_fasta) or os.path.getsize(gen_fasta) < 1:
        require(gen_fasta_org)
        galgo(r'''fa_select {fasta} -n $(cat {list} | awk '{{print "HLA:" $1}}' | tr '\n' ' ') > {out_fasta}'''
                .format(fasta=gen_fasta_org, list=allele_list, out_fasta=gen_fasta),
                stdout=2)

    if args.force_rerun or not os.path.exists(gen_blastdb_nhr):
        require(gen_fasta)
        sh.call('makeblastdb -in {fasta} -out {db} -dbtype nucl -parse_seqids'.format(fasta=gen_fasta, db=gen_blastdb))
        #sh.call_if('', 'wc -l {fasta}')

    if args.force_rerun or not os.path.exists(contig_gen_blast):
        require(gen_blastdb_nhr)
        sh.call("""blastn -db {db} -num_threads {njobs} -query {fasta} -evalue 0.1 -outfmt '7 std gaps qlen slen qseq sseq'"""
            .format(db=gen_blastdb, fasta=contigs_fasta, njobs=args.njobs),
            stdout=contig_gen_blast)

    #@after(indexing)
    if args.force_rerun or not os.path.exists(contig_gen_blast_bests):
        require(contig_gen_blast)
        require(allele_list)
        if not args.dry_run:
            _create_best_contigs(contig_gen_blast, allele_list, contig_gen_blast_bests)

    for gene in args.target_genes:
        nuc = '{dir}/{gene}_nuc.fa'.format(dir=out_dir, gene=gene)
        if args.force_rerun or not os.path.exists(nuc) or os.path.getsize(nuc) < 1:
            galgo(r'''fa_select {org_fasta} -n $(cat {list} | awk '{{print "HLA:" $1}}' | tr '\n' ' ') > {out_fasta}'''
                    .format(
                        org_fasta='{imgt_dir}/fasta/{gene}_nuc.fasta'.format(imgt_dir=imgt_dir, gene=gene),
                        list=allele_list,
                        out_fasta=nuc))

    #    {outdir}/contigs/{contig_name}.fa
    #    {outdir}/contigs/{contig_name}.{gene}.est2genome
    if not args.dry_run:
        require(contig_gen_blast_bests)
        sh.call('mkdir -p {dir}'.format(dir=contig_cdna_dir))
        env = {'out_dir': out_dir, 'args': args, 'contig_cdna_dir': contig_cdna_dir}   # workaround
        Parallel(n_jobs=args.njobs, backend='threading')(delayed(run_est2genome)(name, gene, env) for name, gene in extract_contig_genes(contig_gen_blast_bests))

    def run_classify_hla():
        require(contig_gen_blast_bests)
        require(allele_list)
        def find_closest_exons(name, gene):
            est2genome_parsed = '{dir}/{name}.{gene}_nuc.est2genome.txt'.format(dir=contig_cdna_dir, name=name, gene=gene)
            require(est2genome_parsed)
            tab = pd.read_table(est2genome_parsed)
            tab = tab[tab.category == 'span'].sort_values('edit_distance').sort_values('number', ascending=False).reset_index()
            #logging.info(tab)
            min_edit = tab['edit_distance'].loc[0]
            logging.info((name, gene, min_edit, tab.edit_distance.min()))
            return tab[tab.edit_distance == min_edit]

        def find_full_matched_exons(name, gene):  # TODO check if full length nuc was matched (maybe you have to read nuc lengths from fasta)
            est2genome_parsed = '{dir}/{name}.{gene}_nuc.est2genome.txt'.format(dir=contig_cdna_dir, name=name, gene=gene)
            require(est2genome_parsed)
            tab = pd.read_table(est2genome_parsed)
            tab = tab[(tab.category == 'span')
                    & (tab.number == exon_counts[gene])
                    & (tab.regular_intron == exon_counts[gene] - 1)].sort_values('edit_distance').reset_index()

            #logging.info(tab)
            min_edit = tab['edit_distance'].loc[0]
            logging.info((name, gene, min_edit, tab.edit_distance.min()))
            tab = tab[tab.edit_distance == min_edit]

            # TODO should check aa before edit distance?
            tab['qseq_aa'] = tab.qseq.map(lambda x: dna_translate(x.replace('|', '').replace('-', '')))  # db seq
            tab['rseq_aa'] = tab.rseq.map(lambda x: dna_translate(x.replace('|', '').replace('-', '')))  # contig seq
            tab['aa_match'] = tab['qseq_aa'] == tab['rseq_aa']
            return tab


        contig_genes = list(extract_contig_genes(contig_gen_blast_bests))

        def get_matched_exons():
            #tab = pd.concat(find_closest_exons(name, gene) for name, gene in contig_genes)
            tab = pd.concat(find_full_matched_exons(name, gene) for name, gene in contig_genes)
            return tab

        def get_top_genomes():
            tab1 = pd.read_table(contig_gen_blast_bests)
            #tab = pd.concat(find_closest_exons(name, gene) for name, gene in extract_contig_genes(contig_gen_blast_bests))
            def gen_tops():
                for sname, tab in tab1.groupby('qname'):
                    tab = tab.sort_values('edit_distance').reset_index()
                    #logging.info(tab)
                    min_edit = tab['edit_distance'].loc[0]
                    #logging.info(min_edit)
                    yield tab[tab.edit_distance == min_edit]

            tab = pd.concat(gen_tops())
            return tab

        def get_consistent_genomes(qname_subtype_6ds):
            tab1 = pd.read_table(contig_gen_blast_bests)
            #tab = pd.concat(find_closest_exons(name, gene) for name, gene in extract_contig_genes(contig_gen_blast_bests))
            tab1['subtype_6d'] = tab1['subtype'].apply(lambda x: ':'.join(x.split(':')[:3]))

            def gen_tops():
                for qname, tab in tab1.groupby('qname'):
                    subtype_6ds = qname_subtype_6ds.get(qname, [])
                    yield pd.concat(
                            tab[tab['subtype_6d'] == s].sort_values(['edit_distance', 'aln_len'], ascending=[True, False]).nsmallest(1, columns=['edit_distance'])
                            for s in subtype_6ds)

            tab = pd.concat(gen_tops())
            return tab

        tab = get_matched_exons()
        alist = _read_allele_list(allele_list)
        tab = tab.merge(alist, left_on='qname', right_on='id')#, rsuffix='.1')
        tab = tab[tab.columns.difference(['qseq', 'rseq'])] #+ tab[['qseq', 'rseq']]
        tab.sort_values('rname', inplace=True)
        tab.to_csv(closest_exons, sep='\t', index=False)
        logging.info('Create %s', closest_exons)

        # summary contig info
        tab1 = tab[['rname', 'qname', 'subtype', 'edit_distance', 'subtype_6d', 'subtype_4d', 'aa_match']]
        #tab1['subtype_6d'] = tab1['subtype'].apply(lambda x: ':'.join(x.split(':')[:3]))
        #tab1.loc[:, 'subtype_6d'] = tab1['subtype'].apply(lambda x: ':'.join(x.split(':')[:3]))
        #tab1.loc[:, 'subtype_4d'] = tab1['subtype'].apply(lambda x: ':'.join(x.split(':')[:2]))

        contig_info = {}
        # gene
        for name, gene in contig_genes:
            contig_info[name] = {}
            contig_info[name]['gene'] = gene

        # exon
        for name, t in tab1.groupby('rname'):
            contig_info[name]['edit_6d'] = ed = list(t.edit_distance)[0]  # only one for edit 6d
            contig_info[name]['closest_6d'] = closest_6d = list(sorted(t.subtype_6d.unique()))
            contig_info[name]['closest_4d'] = closest_4d = list(sorted(t.subtype_4d.unique()))
            contig_info[name]['aa_match'] = aa_match = [list(t[t.subtype_4d == sub].aa_match)[0] for sub in closest_4d]
            if ed == 0: # exact
                contig_info[name]['exact_6d'] = contig_info[name]['closest_6d'][0]
            if len(aa_match) > 0 and np.any(aa_match): # exact
                idx_match = np.argmax(aa_match)
                contig_info[name]['exact_4d'] = closest_4d[idx_match]

        #tab = get_top_genomes()
        #tab.sort_values('qname', inplace=True)
        #tab.to_csv(closest_genomes, sep='\t', index=False)
        #logging.info('Create %s', closest_genomes)
        contig_closests = dict((name, d['closest_6d']) for name, d in contig_info.items())

        tab = get_consistent_genomes(contig_closests)
        tab.sort_values('qname', inplace=True)
        tab.to_csv(closest_genomes, sep='\t', index=False)
        logging.info('Create %s', closest_genomes)

        #tab1 = tab[['qname', 'subtype', 'edit_distance']]
        tab['exact_6d'] = tab1['subtype'].apply(lambda x: ':'.join(x.split(':')[:3]))

        # genomes
        for name, t in tab.groupby('qname'):
            t = t.sort_values(['edit_distance', 'aln_len'], ascending=[True, False])
            contig_info[name]['edit_8d'] = eds = list(t.edit_distance)
            contig_info[name]['closest_8d'] = list(t.subtype)
            contig_info[name]['ref_left_8d'] = list(t.qlclip)  # reference is query here
            contig_info[name]['ref_right_8d'] = list(t.qrclip)
            contig_info[name]['q2s_edit'] = list(t.q2s_edit)  # reference is query here
            contig_info[name]['s2q_edit'] = list(t.s2q_edit)
            if eds and eds[0] == 0: # exact (not only one, but select one with max aln_len)
                contig_info[name]['exact_8d'] = contig_info[name]['closest_8d'][0]

        # list matched imgt types
        for name, info in contig_info.items():
            if len(info.get('closest_8d', [])):  # if closest subtypes exist, restrict to them
                match_digits = [sub.split(':') for sub in info['closest_8d']]
            else:
                match_digits = [sub.split(':') for sub in info['closest_6d']]
            logging.info('match_digits for %s: %s', name, match_digits)
            is_sub = alist.subtype.map(lambda x: any(x.split(':')[:len(d)] == d for d in match_digits))
            contig_info[name]['subs'] = list(alist[is_sub]['subtype'])
            contig_info[name]['subs_id'] = list(alist[is_sub]['id'])


        # finalize
        def iter_contig_info():   # for pandas
            for name, gene in contig_genes:
                d = contig_info[name]
                yield {
                        'refname': name,
                        'gene': gene,
                        'exact_4d': d.get('exact_4d', ''),
                        'exact_6d': d.get('exact_6d', ''),
                        'exact_8d': d.get('exact_8d', ''),
                        'n_closest_4d': len(d.get('closest_4d', [])),
                        'closest_4d': ','.join(d.get('closest_4d', [])),
                        'n_closest_6d': len(d.get('closest_6d', [])),
                        'closest_6d': ','.join(d.get('closest_6d', [])),
                        'edit_6d': d.get('edit_6d', -1),
                        'n_closest_8d': len(d.get('closest_8d', [])),
                        'closest_8d': ','.join(d.get('closest_8d', [])),
                        'ref_right_8d': ';'.join(map(str, d.get('ref_right_8d', []))),
                        'ref_left_8d': ';'.join(map(str, d.get('ref_left_8d', []))),
                        'edit_8d': ';'.join(map(str, d.get('edit_8d', []))),
                        'edit_8d_q2s': ';'.join(map(str, d.get('q2s_edit', []))),
                        'edit_8d_s2q': ';'.join(map(str, d.get('s2q_edit', []))),
                        'subs': ','.join(d.get('subs', [])),
                        'subs_id': ','.join(d.get('subs_id', [])),
                        }

        tab = pd.DataFrame(iter_contig_info(),
                columns='refname gene exact_4d exact_6d exact_8d n_closest_4d closest_4d n_closest_6d closest_6d edit_6d n_closest_8d closest_8d ref_left_8d ref_right_8d edit_8d edit_8d_q2s edit_8d_s2q subs subs_id'.split(' '))
        tab.to_csv(contig_summary, sep='\t', index=False)

    def run_summary_fasta():
        require(contig_summary)
        require(allele_list)
        tab = pd.read_table(contig_summary)
        imgt6d_refnames = {}  # {imgt_6d: first refnames}
        for row in tab[['refname', 'closest_6d', 'subs']].itertuples():
            if not row.closest_6d:
                continue
            names = row.closest_6d.split(',')
            for name in names:
                imgt6d_refnames.setdefault(name, row.refname)

        alist = _read_allele_list(allele_list)
        subtype_hla_ids = defaultdict(lambda : defaultdict(list))   # e.g. {'A': {'A*01:01:01': ['HLA00001', 'HLA02169', 'HLA14798']}}
        hla_id_set = set([sid for subs in tab.subs_id if isinstance(subs, basestring) for sid in subs.split(',')])  # only create fasta from contigs in subs_id
        for row in alist[['id', 'gene', 'subtype_6d']].itertuples():
            if row.id in hla_id_set: #row.subtype_6d in imgt6d_refnames and row.id in hla_id_set:
                subtype_hla_ids[row.gene][row.subtype_6d].append(row.id)

        # create gen file (contig name is like HLA:HLA00001)
        for gene in args.target_genes:
            fasta_ids = ['HLA:' + id for ids in subtype_hla_ids.get(gene, {}).values() for id in ids]
            out_fasta = contig_gen_fa_tmpl.format(gene=gene)
            if not fasta_ids:
                continue
            galgo('fa_select {fasta} -n {names} > {out_fasta}'.format(names=' '.join(fasta_ids), fasta=gen_fasta, out_fasta=out_fasta))

        # create nuc file (contig name is like HLA00001)
        for gene in args.target_genes:
            sh.call(': > {out_fasta}'.format(out_fasta=contig_nuc_fa_tmpl.format(gene=gene)))

        for subtype_6d, refname in imgt6d_refnames.items():
            gene = subtype_6d[0]  # A, B, C, H
            est_fasta = '{dir}/{name}.{gene}_nuc.est2genome.fa'.format(dir=contig_cdna_dir, name=refname, gene=gene)
            out_fasta = contig_nuc_fa_tmpl.format(gene=gene)
            hla_ids = subtype_hla_ids[gene][subtype_6d]
            galgo('fa_select {fasta} -n {hla_ids} >> {out_fasta}'.format(fasta=est_fasta, hla_ids=' '.join(hla_ids), out_fasta=out_fasta))

    if not args.dry_run:
        run_classify_hla()
        run_summary_fasta()


class IMGTHLANamer:
    """
    >>> namer = IMGTHLANamer()
    >>> namer.gen_name('A')
    'A*x1'
    >>> namer.gen_name('A')
    'A*x2'
    >>> namer.gen_name('A', 'A*01')
    'A*01:x1'
    >>> namer.gen_name('B')
    'B*x1'
    """

    def __init__(self):
        self._offsets = defaultdict(int)  # A*01:01 : 1
        self._prefix = 'x'

    def gen_name(self, gene, subtype=None):
        key = (gene, subtype)
        self._offsets[key] += 1
        last = self._prefix + str(self._offsets[key])
        if subtype is None:
            return gene + '*' + last
        else:
            return subtype + ':' + last


def isnan(obj):
    return isinstance(obj, float) and np.isnan(obj)

@command.add_sub
@argument('contig_summary')
def imgt_name(args):
    tab = pd.read_table(args.contig_summary)
    namer = IMGTHLANamer()
    print ('refname', 'imgt_name', sep='\t')
    for row in tab.itertuples():
        gene = row.gene
        subtype = None
        if not isnan(row.exact_4d):
            subtype = row.exact_4d
        if not isnan(row.exact_6d):
            subtype = row.exact_6d
        if not isnan(row.exact_8d):
            subtype = row.exact_8d
        name = namer.gen_name(gene, subtype)
        print (row.refname, name, sep='\t')


@command.add_sub
@argument('allele_list')
@argument('subtype', help='e.g. A*01:01:01')
def imgt_subtypes(args):
    require(args.allele_list)
    alist = _read_allele_list(args.allele_list)
    digits = args.subtype.split(':')
    for row in alist[['id', 'subtype']].itertuples():
        if row.subtype.split(':')[:len(digits)] == digits:  # this is 
            print (row.id, row.subtype, sep='\t')


def galgo(args, **kwds):
    import sys
    if isinstance(args, basestring):
        args = ' '.join((sys.executable, sys.argv[0], args))
    else:
        args = [sys.executable, sys.argv[0]] + list(args)
    return sh.call(args, **kwds)



def run_est2genome(name, gene, env):
    out_dir = env['out_dir']
    contig_cdna_dir = env['contig_cdna_dir']
    args = env['args']
    contigs_fasta = args.contigs
    nuc = '{dir}/{gene}_nuc.fa'.format(dir=out_dir, gene=gene)
    contig = '{dir}/{name}.fa'.format(dir=contig_cdna_dir, name=name)
    est2genome = '{dir}/{name}.{gene}_nuc.est2genome'.format(dir=contig_cdna_dir, name=name, gene=gene)
    est2genome_parsed = '{dir}/{name}.{gene}_nuc.est2genome.txt'.format(dir=contig_cdna_dir, name=name, gene=gene)
    est2genome_fasta = '{dir}/{name}.{gene}_nuc.est2genome.fa'.format(dir=contig_cdna_dir, name=name, gene=gene)
    logging.info('target: %s', est2genome)
    if args.force_rerun or not os.path.exists(est2genome) or os.path.getsize(est2genome) < 1:
        require(contigs_fasta)
        galgo('fa_select {fasta} -n {name} > {contig}'.format(fasta=contigs_fasta, name=name, contig=contig))
        require(contig)
        sh.call('''
        {EST2GENOME} -estsequence {nuc} -genomesequence {contig} -intronpenalty 40 -splicepenalty 1 -mismatch 1 -align 1 -usesplice -outfile {outfile}
        '''.format(EST2GENOME=args.EST2GENOME, nuc=nuc, contig=contig, outfile=est2genome))

    if args.force_rerun or not os.path.exists(est2genome_parsed) or os.path.getsize(est2genome_parsed) < 1:
        require(est2genome)
        galgo('''est2genome_parse {est2genome} > {out}'''.format(est2genome=est2genome, out=est2genome_parsed))

    if args.force_rerun or not os.path.exists(est2genome_fasta) or os.path.getsize(est2genome_fasta) < 1:
        require(est2genome)
        galgo('''est2genome_parse {est2genome} -f fasta > {out}'''.format(est2genome=est2genome, out=est2genome_fasta))

def extract_contig_genes(blast_bests):
    tab = pd.read_table(blast_bests)
    print (tab[tab['rank'] == 0])
    for row in tab[tab['rank'] == 0][['qname', 'gene']].itertuples():
        yield row.qname, row.gene


def _read_allele_list(allele_list, fasta_prefix='HLA:'):
    alist = pd.read_table(allele_list, sep=' ', names=['id', 'subtype'])
    alist['fasta_id'] = fasta_prefix + alist['id']
    alist['gene'] = alist.subtype.map(lambda x: x.split('*')[0])
    alist['subtype_8d'] = alist.subtype.map(lambda x: ':'.join(x.split(':')[:4]))
    alist['subtype_6d'] = alist.subtype.map(lambda x: ':'.join(x.split(':')[:3]))
    alist['subtype_4d'] = alist.subtype.map(lambda x: ':'.join(x.split(':')[:2]))
    alist['subtype_2d'] = alist.subtype.map(lambda x: ':'.join(x.split(':')[:1]))
    return alist

def _create_best_contigs(blastn, allele_list, bests):
    keys = 'qname rank sname gene subtype aln_len qlen slen identity edit_distance mismatches gaps slclip srclip qlclip qrclip left_overhang right_overhang s2q_edit q2s_edit'.split(' ')
    tab = _get_best_contigs(blastn)
    alist = _read_allele_list(allele_list)
    tab = tab.merge(alist, left_on='sname', right_on='fasta_id')#, rsuffix='.1')
    sseq = np.where(tab.is_reverse, tab.sseq.apply(dna_revcomp), tab.sseq)
    sstart = np.where(tab.is_reverse, tab.send, tab.sstart)
    def encode_edit(edit_info):
        return ','.join(('{0}:{1}>{2}'.format(*e) for e in edit_info))
    tab = tab[keys]
    tab.loc[:, 's2q_edit'] = [encode_edit(calc_edit_info(q, s, start)) for q, s, start in zip(tab.qseq, sseq, sstart)]
    tab.loc[:, 'q2s_edit'] = [encode_edit(calc_edit_info(s, q, start)) for s, q, start in zip(sseq, tab.qseq, tab.qstart)]

    #tab.to_csv('/dev/stdout', sep='\t', index=False)
    tab.to_csv(bests, sep='\t', index=False)


def _get_best_contigs(blastn):
    min_sratio = .95
    max_overhang = 20
    #max_edit_distance = 20
    max_edit_distance = 50

    require(blastn)

    with open(blastn) as reader: #, open(summary, 'w+') as writer:
        blast = BlastTabFile(reader)
        def iter_tabs():
            for query, tab in blast.iter_query_tables():
                match_cond = ((tab.sratio >= min_sratio)
                    & (tab.left_overhang <= max_overhang)
                    & (tab.right_overhang <= max_overhang)
                    & (tab.edit_distance <= max_edit_distance))
                tab = tab[match_cond]
                tab = tab.sort_values(['edit_distance', 'aln_len'], ascending=[True, False])
                tab.reset_index(drop=True, inplace=True)
                tab.loc[:, 'rank'] = tab.index
                yield tab

        return pd.concat(iter_tabs()) #ignore_index=True)


def calc_edit_info(query_aln, ref_aln, ref_start=1): #TODO simplify
    """
    #        123 4567   890123456 789012 # offset
    >>> q = 'ACGTCGTTTGG--GACC-GTTTATTCA'
    >>> r = 'AAG-CGTT---ATGTCCAGT-TGTTAA'
    >>> l = list(calc_edit_info(q, r, 3001))
    >>> l[0]
    (3002, 'A', 'C')
    >>> l[1]
    (3004, '', 'T')
    >>> l[2]
    (3008, 'AT', 'TGG')
    >>> l[3]
    (3011, 'T', 'A')
    >>> l[4]
    (3014, 'A', '')
    >>> l[5]
    (3017, '', 'T')
    >>> l[6]
    (3018, 'G', 'A')
    >>> l[7]
    (3021, 'A', 'C')
    """
    assert len(query_aln) == len(ref_aln)
    edits = []
    offset = 0
    apos = ref_start
    for r, q in zip(ref_aln, query_aln):
        if r != q:
            edits.append((apos, offset, r.replace('-', ''), q.replace('-', '')))
        if r == '-':
            offset += 1
        apos += 1

    prev = -float('inf')
    outs = []
    for apos, offset, r, q in edits:
        if apos - prev == 1:
            outs[-1]['ref'] += r
            outs[-1]['query'] += q
        else:
            outs.append({'pos': apos - offset, 'ref': r, 'query': q})
        prev = apos

    return [(out['pos'], out['ref'], out['query']) for out in outs]


def require(file):
    if not os.path.exists(file):
        raise Exception('File does not exist: {0}'.format(file))


@command.add_sub
@argument('est2genome')
@argument('-f', '--format', choices=['tab', 'fasta'], default='tab', help='output format')
def est2genome_parse(args):
    """
    span
        alignment bases
        '|': intron
        # TODO check intron category (regular or irregular)
    """
    with open(args.est2genome) as fp:
        aln_keys = ['edit_distance', 'mismatches', 'gaps', 'qseq', 'rseq']
        if args.format == 'tab':
            print (*('qname qstart qend category number rname rstart rend score match_rate regular_intron'.split(' ') + aln_keys), sep='\t')
            for rec in Est2GenomeRecord.parse(fp):
                aln = [getattr(rec.span, k) for k in aln_keys]
                regular_intron = len([i for i in rec.introns if i.type in '+-'])
                print (rec.qname, rec.span.qstart, rec.span.qend, 'span', len(rec.exons), rec.span.rname, rec.span.rstart, rec.span.rend, rec.span.score, rec.span.match_rate, regular_intron, *aln, sep='\t')
                for num, rec in enumerate(rec.exons, 1):
                    aln = [getattr(rec, k) for k in aln_keys]
                    print (rec.qname, rec.qstart, rec.qend, 'exon', num, rec.rname, rec.rstart, rec.rend, rec.score, rec.match_rate, 0, *aln, sep='\t')
        elif args.format == 'fasta':
            for rec in Est2GenomeRecord.parse(fp):
                print ('>' + rec.qname)
                seqs = []
                last_rend = None
                for num, rec in enumerate(rec.exons, 1):
                    if last_rend is not None:
                        fill_len = rec.rstart - last_rend
                        seqs.append('N' * fill_len)
                    seqs.append(rec.qseq)
                    last_rend = rec.rend
                print (''.join(seqs))



class Est2GenomeRecord:
    #Span = namedtuple('Span', 'score match_rate rstart rend rname qstart qend qname rseq qseq')
    #Exon = namedtuple('Exon', 'score match_rate rstart rend rname qstart qend qname rseq qseq')
    class Feature(namedtuple('Feature', 'score match_rate rstart rend rname qstart qend qname rseq qseq')):
        def __init__(self, *args, **kwds):
            super(Est2GenomeRecord.Feature, self).__init__(*args, **kwds)
            self.mismatches = self.gaps = self.edit_distance = None
            if self.rseq is not None:
                self.mismatches = self.gaps = 0
                for r, q in zip(self.rseq, self.qseq):
                    if r == '|':
                        continue
                    elif r == '-' or q == '-':
                        self.gaps += 1
                    elif r != q:
                        self.mismatches += 1
                self.edit_distance = self.mismatches + self.gaps

    Intron = namedtuple('Intron', 'type score match_rate rstart rend rname')

    """
    """

    _intron_pat = re.compile(r'[\.a-z]+')   # split at 'CGAGGCCGgtgag.....gccagGGTCTCACACTTGGCAGACGATGTATG'

    @staticmethod
    def _is_alignment_start(line):
        tokens = blank_split(line)
        return len(tokens) == 3 and tokens[1] == 'vs' and tokens[2][-1] == ':'

    @staticmethod
    def _is_new_block(line):
        return line.startswith('Note')

    @classmethod
    def parse(cls, fp):
        try:
            it = fp
            #it = dropwhile(lambda x: not _is_new_block(x), it)  # find a line starts with 'Note'
            line, it = skip_until(cls._is_new_block, it)  # find new block
            # body
            while 1:
                splice_lines, it = collect_while(make_not(isblank), it)  # collect exon, intron records
                next(it)  # blank
                span_line = next(it)  # blank
                next(it)  # blank
                segment_lines, it = collect_while(make_not(isblank), it)

                splicings = [blank_split(line.rstrip()) for line in splice_lines]
                segments = [blank_split(line.rstrip()) for line in segment_lines]
                span_row = blank_split(span_line)

                rname = span_row[5]
                qname = span_row[8]
                logging.debug('%s vs %s', rname, qname)

                # find alignment or next block
                line, it = skip_until(lambda x: cls._is_alignment_start(x) or cls._is_new_block(x), it)
                rseqs = qseqs = None
                if cls._is_alignment_start(line):
                    next(it)  # skip blank line
                    rbuf = []
                    qbuf = []
                    while 1:  # reading alignment
                        tokens = blank_split(next(it))  # ref line
                        if not tokens or tokens[0] != rname:  # end of alignment block
                            break
                        rbuf.append(tokens[2])
                        next(it)  # match line
                        tokens = blank_split(next(it))  # query line
                        qbuf.append(tokens[2])
                        next(it)  # blank line

                    rseqs = cls._intron_pat.split(''.join(rbuf))
                    qseqs = cls._intron_pat.split(''.join(qbuf))

                yield cls(span_row, splicings, segments, rseqs, qseqs)
                skip_until(cls._is_new_block, it)
        except StopIteration as e:
            pass

    def __init__(self, span_row, splice_rows, segment_rows, rseqs, qseqs):
        row = span_row
        if rseqs is not None:
            rseq = '|'.join(rseqs)
            qseq = '|'.join(qseqs)
        else:
            rseq = qseq = None
        self.span = self.Feature(int(row[1]), float(row[2]), int(row[3])-1, int(row[4]), row[5], int(row[6])-1, int(row[7]), row[8], rseq, qseq)
        self.exons = []
        self.introns = []
        exon_i = 0
        for row in splice_rows:
            if row[0] == 'Exon':
                if rseqs is not None:
                    rseq = rseqs[exon_i]
                    qseq = qseqs[exon_i]
                else:
                    rseq = qseq = None
                rec = self.Feature(int(row[1]), float(row[2]), int(row[3])-1, int(row[4]), row[5], int(row[6])-1, int(row[7]), row[8], rseq, qseq)
                self.exons.append(rec)
                exon_i += 1
            elif row[0][1:] == 'Intron':
                intron_type = row[0][0] # (+, -, ?)
                rec = self.Intron(intron_type, int(row[1]), float(row[2]), int(row[3])-1, int(row[4]), row[5])
                self.introns.append(rec)
            else:
                logging.error(row)
                raise NotImplementedError

        self.qname = self.span.qname
        self.rname = self.span.rname
        #self.segment_rows = segment_rows  # TODO



def read_vbseq(result, mean_rlen):
    """
    ID      LENGTH  Z       FPKM    THETA
    HLA:HLA00001    1047    0.00    0.0     0.000000e+00
    HLA:HLA00002    1047    0.00    0.0     0.000000e+00
    HLA:HLA00398    1047    0.00    0.0     0.000000e+00
    HLA:HLA00644    1047    0.00    0.0     0.000000e+00
    HLA:HLA00003    1047    45.74   7877.1781230    1.456456e-06
    """
    tab = pd.read_table(result)
    tab['mean_depth'] = 1. * mean_rlen * tab['Z'] / tab['LENGTH']
    return tab


@command.add_sub
@argument('vbseq')
@argument('-a', '--allele-list', required=True, help='IMGT HLA Allelelist')
@argument('--fasta-prefix', default='HLA:', help='HLA prefix of Allelelist')
@argument('-r', '--mean-rlen', required=True, type=float, help='mean read length (if using paired-end mode, set twice of the single read length)')
@argument('--min-depth', type=float, default=0, help='min read depth (only used for output)')
@argument('--min-ratio', type=float, default=0, help='min read ratio (only used for output)')
@argument('--min-het-ratio', type=float, default=.25)
@argument('--min-hom-ratio', type=float, default=.75)
@argument('--hom-x', type=float, default=2.5)
@argument('-g', '--genes', nargs='+')
@argument('-c', '--copy-numbers', nargs='+', type=int)
@argument('-f', '--format', choices=['v0', 'v1'], default='v1', help='version of output format')
@argument('-s', '--sampleid')
def vbseq_hla_gt(args):
    """
    """
    require(args.allele_list)
    logging.info('a prefix of refname is: %s', args.fasta_prefix)
    alist = _read_allele_list(args.allele_list, fasta_prefix=args.fasta_prefix)
    default_cn = 2
    gene_cns = {}
    if args.copy_numbers:
        assert len(args.genes) == len(args.copy_numbers)
        gene_cns = dict(zip(args.genes, args.copy_numbers))

    tab = read_vbseq(args.vbseq, args.mean_rlen)
    tab = tab.merge(alist, left_on='ID', right_on='fasta_id')#, rsuffix='.1')
    if args.genes:
        gene_set = set(args.genes)
        tab = tab[tab.gene.map(lambda x: x in gene_set)]

    del tab['fasta_id']
    del tab['id']
    gene_depths = {}
    for gene, gtab in tab.groupby('gene'):
        gene_depths[gene] = gtab.mean_depth.sum()

    keys = '8d 6d 4d 2d'.split(' ')
    depths = {key: {} for key in keys}
    for key in keys:
        for gkey, gtab in tab.groupby('subtype_' + key):
            depths[key][gkey] = gtab.mean_depth.sum()

    tab['gene_cn'] = tab.gene.map(lambda x: gene_cns.get(x, default_cn))
    tab['gene_depth'] = tab.gene.map(lambda x: gene_depths[x])
    tab['ratio_in_gene'] = tab.mean_depth / tab.gene_depth
    for key in keys:
        d = tab['subtype_' + key].map(lambda x: depths[key][x])
        tab['ratio_' + key] = d / tab.gene_depth

    # add genotype
    def add_gt(tab, id_key='ID', ratio_key='ratio_in_gene', gt_key='gt', sel_key='sel'):
        def get_2copy(gtab):
            gtab = gtab.drop_duplicates(id_key, keep='first')
            tops = gtab[gtab[ratio_key] >= args.min_het_ratio].sort_values(ratio_key, ascending=False)[:2]
            if len(tops) >= 2:
                if tops.iloc[0][ratio_key] >= args.min_hom_ratio:
                    if tops.iloc[0][ratio_key] > tops.iloc[1][ratio_key] * args.hom_x:
                        return ('hom', [tops.iloc[0][id_key]])
                    else:
                        return ('het', [tops.iloc[0][id_key], tops.iloc[1][id_key]])
                else:
                    return ('het', [tops.iloc[0][id_key], tops.iloc[1][id_key]])
            elif len(tops) == 1:
                if tops.iloc[0][ratio_key] >= args.min_hom_ratio:
                    return ('hom', [tops.iloc[0][id_key]])
                else:
                    return ('het', [tops.iloc[0][id_key]])
            else:
                return ('unknown', [])

        def get_1copy(gtab):
            gtab = gtab.drop_duplicates(id_key, keep='first')
            tops = gtab[gtab[ratio_key] >= args.min_hom_ratio].sort_values(ratio_key, ascending=False)[:2]
            if len(tops) >= 1:
                return ('hom', [tops.iloc[0][id_key]])
            else:
                return ('unknown', [])

        gene_gts = {}
        selected_ids = set()
        for gene, gtab in tab.groupby('gene'):
            cn = gtab.iloc[0]['gene_cn']
            if cn == 2:
                gt, ids = get_2copy(gtab)
            elif cn == 1:
                gt, ids = get_1copy(gtab)
            elif cn == 0:
                gt, ids = 'hom', []
            else:
                raise NotImplemented
            gene_gts[gene] = gt
            selected_ids.update(set(ids))

        tab[gt_key] = tab.gene.map(lambda x: gene_gts[x])
        tab[sel_key] = tab[id_key].map(lambda x: int(x in selected_ids))

    add_gt(tab)
    for key in keys:
        add_gt(tab, id_key='subtype_' + key, ratio_key='ratio_' + key, gt_key='gt_' + key, sel_key='sel_' + key)

    #print (tab)
    cond_show = (tab['ratio_in_gene'] >= args.min_ratio) & (tab['mean_depth'] >= args.min_depth)
    tab = tab[cond_show]
    tab = tab.sort_values('gene')
    if args.format == 'v0':
        pass
    if args.format == 'v1':
        tab = convert_hla_gt_format(tab, sampleid=args.sampleid)
    tab.to_csv('/dev/stdout', index=False, sep='\t')


def convert_hla_gt_format(org_tab, sampleid='sampleid'):
    """ # TODO
    emit untyped record?
    emit copy number 0 record?
    """
    digit = 0
    tabs = []
    cols = 'sampleid gene digit subtype gene_cn allele_cn allele sel gene_depth ratio refnames'.split(' ')
    digits = [2, 4, 6, 8, 0]

    for gene, tab1 in org_tab.groupby('gene'):
        for digit in digits:
            if digit == 0:
                id_key = 'ID'
                ratio_key = 'ratio_in_gene'
                gt_key = 'gt'
                sel_key = 'sel'
            else:
                id_key = 'subtype_{0}d'.format(digit)
                ratio_key = 'ratio_{0}d'.format(digit)
                gt_key = 'gt_{0}d'.format(digit)
                sel_key = 'sel_{0}d'.format(digit)

            rtab = tab1.groupby(id_key).agg({'ID': lambda x: ','.join(x)}).rename(columns={'ID': 'refnames'}).reset_index()
            tab2 = tab1.drop_duplicates(id_key, keep='first')
            tab2 = tab2.merge(rtab)
            tab2 = tab2.sort_values(ratio_key, ascending=False).reset_index()
            recs = {col: [] for col in cols}
            for an, rec in enumerate(tab2.itertuples(), 1):
                recs['sampleid'].append(sampleid)
                recs['gene'].append(gene)
                recs['digit'].append(digit)
                recs['gene_cn'].append(rec.gene_cn)
                recs['gene_depth'].append(rec.gene_depth)
                recs['allele'].append(an)
                gt = getattr(rec, gt_key)
                if rec.gene_cn == 0:
                    acn = 0
                elif rec.gene_cn == 1 and gt == 'hom':
                    acn = 1
                elif rec.gene_cn == 2 and gt == 'hom':
                    acn = 2
                elif rec.gene_cn == 2 and gt == 'het':
                    acn = 1
                else:
                    acn = 0
                sel = getattr(rec, sel_key)

                recs['allele_cn'].append(acn * sel)
                recs['subtype'].append(getattr(rec, id_key))
                recs['sel'].append(sel)
                recs['ratio'].append(getattr(rec, ratio_key))
                recs['refnames'].append(rec.refnames)
            tab_out = pd.DataFrame(recs, columns=cols)
            tabs.append(tab_out)

    tab = pd.concat(tabs)[cols]
    return tab


@command.add_sub
@argument('hla_gt_table')
@argument('-s', '--sampleid')
@argument('--min-depth', type=float, default=0, help='min read depth (only used for output)')
@argument('--min-ratio', type=float, default=0, help='min read ratio (only used for output)')
def convert_hla_gt(args):
    """ Update format
    """
    tab = pd.read_table(args.hla_gt_table)
    tab = convert_hla_gt_format(tab, sampleid=args.sampleid)
    cond_show = (tab['ratio_in_gene'] >= args.min_ratio) & (tab['mean_depth'] >= args.min_depth)
    tab = tab[cond_show]
    tab.to_csv('/dev/stdout', index=False, sep='\t')





def load_gender_info(gender_info):
    sample_genders = {}
    with open(gender_info) as fp:
        for line in fp:
            sample, gender = line.rstrip().split()
            gender = int(gender)
            assert gender in (1, 2)
            sample_genders[sample] = gender

    return sample_genders

def get_ploidies(chrom, genders):
    if chrom in ('X', 'chrX'):
        return tuple(1 if s == 1 else 2 for s in genders)
    elif chrom in ('Y', 'chrY'):
        return tuple(1 if s == 1 else 0 for s in genders)
    return (2,) * len(genders)
